from os import path
import pytest
import numpy as np


from tests.python.helpers.utils import verify_points, basic_polylidar_verification, verify_all_polygons_are_valid, verify_all


np.random.seed(1)

def test_random_points(random_points, params_lmax):
    verify_all(random_points, params_lmax)

def test_clusters(cluster_groups):
    points = cluster_groups['points']
    polylidar_kwargs = cluster_groups['polylidar_kwargs']
    verify_all(points, polylidar_kwargs)





try:
    from .polylidar_pybind import * # py2 py3 compatible
except Exception:
    # this was installed with as a python wheel
    from polylidar_pybind import *

try:
    import pkg_resources  # part of setuptools
    __version__ = pkg_resources.require("polylidar")[0].version
except Exception:
    __version__ = '@PROJECT_VERSION@'



"""
Demonstrates Polygon Extraction on some RealSense data.
Demonstrates unorganized and organized point clouds polygon extraction
Requires: Fixtures data
"""
import time
import logging
import sys

import numpy as np

from examples.python.util.realsense_util import (get_realsense_data, get_frame_data, R_Standard_d400, prep_mesh,
                                                 create_open3d_pc, extract_mesh_planes, COLOR_PALETTE)

from polylidar import (Polylidar3D, MatrixDouble, MatrixFloat, extract_tri_mesh_from_float_depth,
                       extract_point_cloud_from_float_depth)

from polylidar.polylidarutil.open3d_util import construct_grid, create_lines, flatten, create_open_3d_mesh_from_tri_mesh
from polylidar.polylidarutil.plane_filtering import filter_planes_and_holes
from polylidar.polylidarutil.line_mesh import o3d_major_version

import open3d as o3d


def filter_and_create_open3d_polygons(points, polygons):
    " Apply polygon filtering algorithm, return Open3D Mesh Lines "
    config_pp = dict(filter=dict(hole_area=dict(min=0.025, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.5)),
                     positive_buffer=0.00, negative_buffer=0.02, simplify=0.01)
    planes, obstacles = filter_planes_and_holes(polygons, points, config_pp)
    all_poly_lines = create_lines(planes, obstacles, line_radius=0.01)
    return all_poly_lines


def run_test(pcd, rgbd, intrinsics, extrinsics, bp_alg=dict(radii=[0.02, 0.02]), poisson=dict(depth=8), callback=None, stride=2):
    """Demonstrate Polygon Extraction on both unorganized and organized point cloud

    Args:
        pcd (o3d.geometry.PointCloud): Open3D point cloud
        rgbd (np.ndarray): MXN Numpy array
        intrinsics (np.ndarray): 3X3 numpy array of camera intrinsics (assume pin hole model)
        extrinsics (np.ndarray): 4X4 numpy array of extrinsics of camera
        bp_alg (dict, optional): Arguments to Open3D ball pivot alg. Defaults to dict(radii=[0.02, 0.02]).
        poisson (dict, optional): Arguments to Open3D Poisson surface reconstruction. Defaults to dict(depth=8).
        callback (function, optional): Callback function for visualization. Defaults to None.
        stride (int, optional): Skip rows/columns in rgbd. Defaults to 2.

    """
    points = np.asarray(pcd.points)
    polylidar_kwargs = dict(alpha=0.0, lmax=0.10, min_triangles=100,
                            z_thresh=0.04, norm_thresh=0.90, norm_thresh_min=0.90, min_hole_vertices=6)
    pl = Polylidar3D(**polylidar_kwargs)

    ################################################################################
    ##### Treat data as an unorganized point clouds
    ##### Create Surface Mesh using 2.5 Delaunay Triangulation and extract Polygons
    ################################################################################
    points_mat = MatrixDouble(points)
    t1 = time.perf_counter()
    mesh, planes, polygons = pl.extract_planes_and_polygons(points_mat)
    t2 = time.perf_counter()

    # Visualization of mesh and polygons
    all_poly_lines = filter_and_create_open3d_polygons(points, polygons)
    triangles = np.asarray(mesh.triangles)
    mesh_2d_polylidar = extract_mesh_planes(points, triangles, planes, mesh.counter_clock_wise, COLOR_PALETTE[0])
    mesh_2d_polylidar.extend(flatten([line_mesh.cylinder_segments for line_mesh in all_poly_lines]))
    time_mesh_2d_polylidar = (t2 - t1) * 1000
    polylidar_alg_name = 'Treated as **Unorganized** Point Cloud - 2.5D Delaunay Triangulation with Polygon Extraction'
    callback(polylidar_alg_name, time_mesh_2d_polylidar, pcd, mesh_2d_polylidar)

    ################################################################################
    ###### Treat data as an **Organized** 3D Point Cloud #########
    ###### Creates a true 3D mesh and is much master using organized structure of point cloud 
    ################################################################################
    tri_mesh, t_mesh_creation = make_uniform_grid_mesh(np.asarray(
        rgbd.depth), np.ascontiguousarray(intrinsics.intrinsic_matrix), extrinsics, stride=stride)

    # Visualization of only the mesh
    tri_mesh_o3d = create_open_3d_mesh_from_tri_mesh(tri_mesh)
    uniform_alg_name = 'Treated as **Organized** Point Cloud - Right-Cut Triangulation/Uniform Mesh (Mesh only)'
    callback(uniform_alg_name, t_mesh_creation, pcd, tri_mesh_o3d)

    # Exctact Polygons with Polylidar3D using the Uniform Mesh. Dominant Plane normal is 0,0,1.
    t1 = time.perf_counter()
    planes, polygons = pl.extract_planes_and_polygons(tri_mesh)
    t2 = time.perf_counter()

    # Visualization of mesh and polygons
    vertices_np = np.asarray(tri_mesh.vertices)
    triangles_np = np.asarray(tri_mesh.triangles)

    all_poly_lines = filter_and_create_open3d_polygons(vertices_np, polygons)
    mesh_3d_polylidar = extract_mesh_planes(vertices_np, triangles_np, planes, tri_mesh.counter_clock_wise)
    mesh_3d_polylidar.extend(flatten([line_mesh.cylinder_segments for line_mesh in all_poly_lines]))
    time_polylidar3D = (t2 - t1) * 1000
    polylidar_3d_alg_name = 'Polygon Extraction on Uniform Mesh (only one dominant plane normal)'
    callback(polylidar_3d_alg_name, time_polylidar3D,
             create_open3d_pc(vertices_np), mesh_3d_polylidar)

    ##### Uncomment if you are interested in other mesh creation techniques #####

    # # Estimate Point Cloud Normals
    # t3 = time.perf_counter()
    # pcd.estimate_normals(
    #     search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.10, max_nn=20))
    # t4 = time.perf_counter()
    # time_estimate_point_normals = (t4 - t3) * 1000
    # point_normal_alg_name = 'Point Normal Estimation'
    # callback(point_normal_alg_name, time_estimate_point_normals, pcd, None)
    # # Create True 3D Surface Mesh using Ball Pivot Algorithm
    # radii = o3d.utility.DoubleVector(bp_alg['radii'])
    # t5 = time.perf_counter()
    # mesh_ball_pivot = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
    #     pcd, radii)
    # prep_mesh(mesh_ball_pivot)
    # t6 = time.perf_counter()
    # time_mesh_ball_pivot = (t6 - t5) * 1000
    # ball_point_alg_name = 'Ball Pivot'
    # callback(ball_point_alg_name, time_mesh_ball_pivot, pcd, mesh_ball_pivot)
    # # Create True 3D Surface Mesh using Poisson Reconstruction Algorithm
    # t7 = time.perf_counter()
    # mesh_poisson, densities = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(
    #     pcd, **poisson)
    # vertices_to_remove = densities < np.quantile(densities, 0.1)
    # mesh_poisson.remove_vertices_by_mask(vertices_to_remove)
    # t8 = time.perf_counter()
    # prep_mesh(mesh_poisson)
    # time_mesh_poisson = (t8 - t7) * 1000
    # poisson_alg_name = 'Poisson'
    # callback(poisson_alg_name, time_mesh_poisson, pcd, mesh_poisson)

    # results = [
    #     dict(alg=polylidar_alg_name, mesh=mesh_2d_polylidar,
    #          execution_time=time_mesh_2d_polylidar),
    #     dict(alg=point_normal_alg_name, mesh=None,
    #          execution_time=time_estimate_point_normals),
    #     dict(alg=ball_point_alg_name, mesh=mesh_ball_pivot,
    #          execution_time=time_mesh_ball_pivot),
    #     dict(alg=poisson_alg_name, mesh=mesh_poisson,
    #          execution_time=time_mesh_poisson)
    # ]
    # return results


def make_uniform_grid_mesh(im, intrinsics, extrinsics, stride=2, **kwargs):
    """Create a Unifrom Grid Mesh from an RGBD Image

    Arguments:
        img {ndarray} -- MXN Float Depth Image
        intrinsics {ndarray} -- 3X3 intrinsics matrix
        extrinsics {ndarray} -- 4X4 matrix

    Keyword Arguments:
        stride {int} -- Stride for creating point cloud (default: {2})

    Returns:
        tuple(dict, dict) - Mesh and timings
    """
    t0 = time.perf_counter()
    tri_mesh = extract_tri_mesh_from_float_depth(MatrixFloat(
        im), MatrixDouble(intrinsics), MatrixDouble(extrinsics), stride=stride)
    t1 = time.perf_counter()

    return tri_mesh, (t1-t0) * 1000


def callback(alg_name, execution_time, pcd, mesh=None):
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -0.7])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-1.0, translate=[0, 1.0, 0.0])
    logging.info("%s took %.2f milliseconds", alg_name, execution_time)
    if mesh:
        if isinstance(mesh, list):
            o3d.visualization.draw_geometries(
                [*mesh, pcd, grid_ls, axis_frame])
        else:
            o3d.visualization.draw_geometries([mesh, pcd, grid_ls, axis_frame])


def main():
    color_files, depth_files, traj, intrinsics = get_realsense_data()
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -0.7])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-0.8, translate=[0, 1.0, 0.0])
    for idx in range(len(color_files)):
        if idx < 2:
            continue
        pcd, rgbd, extrinsics = get_frame_data(idx, color_files, depth_files, traj, intrinsics, stride=2)
        center = pcd.get_center() if o3d_major_version > 9 else True
        pcd = pcd.rotate(R_Standard_d400[:3, :3], center=center)

        logging.info("File %r - Point Cloud; Size: %r", idx, np.asarray(pcd.points).shape[0])
        o3d.visualization.draw_geometries([pcd, grid_ls, axis_frame])
        results = run_test(pcd, rgbd, intrinsics, extrinsics, callback=callback, stride=2)


if __name__ == "__main__":
    main()



import time
import math
import sys
from os import path, listdir
from os.path import exists, isfile, join, splitext
import re
import logging

import numpy as np
from polylidar.polylidarutil import COLOR_PALETTE
from polylidar.polylidarutil.line_mesh import o3d_major_version
import open3d as o3d
from scipy.spatial.transform import Rotation as R

DIR_NAME = path.dirname(__file__)
FIXTURES_DIR = path.join(DIR_NAME, '../../..', 'fixtures')
MESHES_DIR = path.join(FIXTURES_DIR, 'meshes')

DENSE_MESH = path.join(MESHES_DIR, 'dense_first_floor_map_smoothed.ply')
SPARSE_MESH = path.join(MESHES_DIR, 'sparse_basement.ply')
BASEMENT_CHAIR = path.join(MESHES_DIR, 'basement_chair_5cm.ply')

ALL_MESHES = [DENSE_MESH, SPARSE_MESH, BASEMENT_CHAIR]
# ALL_MESHES_ROTATIONS = [R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0])),
#                         R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0]))]

ALL_MESHES_ROTATIONS = [None, None, None]


def get_mesh_data_iterator():
    for i, (mesh_fpath, r) in enumerate(zip(ALL_MESHES, ALL_MESHES_ROTATIONS)):
        example_mesh = o3d.io.read_triangle_mesh(str(mesh_fpath))
        if r is not None:
            center = [0, 0, 0] if o3d_major_version > 9 else True
            example_mesh = example_mesh.rotate(r.as_matrix(), center=center)
        example_mesh_filtered = example_mesh
        example_mesh_filtered.compute_triangle_normals()
        yield example_mesh_filtered


def main():
    for i, mesh in enumerate(get_mesh_data_iterator()):
        if i < 1:
            continue
        colors = np.asarray(mesh.vertex_colors)
        colors2 = np.column_stack((colors[:, 2], colors[:, 1], colors[:, 0]))
        mesh.vertex_colors = o3d.utility.Vector3dVector(colors2)
        o3d.io.write_triangle_mesh('test.ply', mesh)

if __name__ == "__main__":
    main()



import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors 
from shapely.geometry import Polygon, MultiPolygon
from .descartes_patch import PolygonPatch

COLOR_PALETTE = list(map(colors.to_rgb, plt.rcParams['axes.prop_cycle'].by_key()['color']))

def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

def apply_rotation(rm, points):
    return (rm @ points.T).T

def get_poly_coords(outline, points):
    return [get_point(pi, points) for pi in outline]

def get_point(pi, points):
    """Get a point from a numpy array.
    If the numpy array has 3 dimensions, will return all three (x,y,z)
    Arguments:
        pi {int} -- Point index in numpy array
        points {ndarray} -- Numpy array of points
    
    Returns:
        [list] -- List of poins [x,y,(z)]
    """
    if points.shape[1] > 2:
        return [points[pi, 0], points[pi, 1], points[pi, 2]]
    else:
        return [points[pi, 0], points[pi, 1]]

def convert_to_shapely_polygons(polygons, points, return_first=False, sort=False, mp=False):
    """Convert Polylidar C++ polygons to shapely polygons
    
    Arguments:
        polygons {vector<polygons>} -- Vector of polygons
        points {ndarray} -- ndarray
    
    Keyword Arguments:
        return_first {bool} -- Return the first polygon (default: {False})
        sort {bool} -- Sort polygons by largest number of vertices  (default: {False})
        mp {bool} -- Convert list of polygons to MultiPolygon (default: {False})
    
    Returns:
        [list or single Polygon or Single Multipolygon] -- Its up to you what you want
    """
    if sort:
        polygons.sort(key=lambda poly: len(poly.shell), reverse=True)

    shapely_polygons = []
    for poly in polygons:
        shell_coords = get_poly_coords(poly.shell, points)
        hole_coords = [get_poly_coords(hole, points) for hole in poly.holes]
        poly_shape = Polygon(shell=shell_coords, holes=hole_coords)
        
        shapely_polygons.append(poly_shape)

    # Return only the largest by number of vertices
    if shapely_polygons and return_first:
        return shapely_polygons[0]

    # Check if a multipolygon
    if len(shapely_polygons) > 1 and mp:
        return MultiPolygon(shapely_polygons)
    else:
        return shapely_polygons

def set_axes_radius(ax, origin, radius, ignore_z=False):
    ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
    ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
    if not ignore_z:
        ax.set_zlim3d([origin[2] - radius, origin[2] + radius])


def set_axes_equal(ax, ignore_z=False):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    set_axes_radius(ax, origin, radius, ignore_z=ignore_z)


def generate_3d_plane(bounds_x=[0,10,0.5], bounds_y=[0, 10, 0.5], holes=[[[3,5], [3, 5]]], height=0, planar_noise=0.01, height_noise=0.1):
    X, Y = np.mgrid[bounds_x[0]:bounds_x[1]:bounds_x[2], bounds_y[0]:bounds_y[1]:bounds_y[2]]
    pc = np.column_stack((X.ravel(), Y.ravel()))

    plane_noise = np.abs(np.random.randn(pc.shape[0], 2)) * planar_noise
    pc[:,:2] = pc[:,:2] + plane_noise

    plane_height = (np.abs(np.random.randn(pc.shape[0])) * height_noise) + height
    pc_noisy = np.column_stack((pc[:,0], pc[:, 1], plane_height))

    mask = np.zeros(pc_noisy.shape[0], dtype=bool)
    for hole in holes:
        holes_x, holes_y = hole
        mask1 = (pc_noisy[:,0] > holes_x[0]) & (pc_noisy[:,0] < holes_x[1])
        mask2 = (pc_noisy[:,1] > holes_y[0]) & (pc_noisy[:,1] < holes_y[1])
        mask3 = mask1 & mask2
        mask = mask | mask3

    return pc_noisy[~mask]


def plot_polygons(polygons, points, ax, linewidth=2, shell_color='green', hole_color='orange'):
    for poly in polygons:
        shell_coords = [get_point(pi, points) for pi in poly.shell]
        outline = Polygon(shell=shell_coords)
        outlinePatch = PolygonPatch(outline, ec=shell_color, fill=False, linewidth=linewidth)
        ax.add_patch(outlinePatch)

        for hole_poly in poly.holes:
            shell_coords = [get_point(pi, points) for pi in hole_poly]
            outline = Polygon(shell=shell_coords)
            outlinePatch = PolygonPatch(outline, ec=hole_color, fill=False, linewidth=linewidth)
            ax.add_patch(outlinePatch)

        
def generate_test_points(num_groups=2, dist=100, group_size=10, seed=1):
    np.random.seed(seed)
    sigma = dist / (num_groups * 2 - 1)
    cov = np.array([[sigma, 0], [0, sigma]])
    a = np.array([0, 0])
    for i in range(0, num_groups):
        mean = np.random.random(2) * dist
        b = np.random.multivariate_normal(mean, cov, size=group_size)
        a = np.vstack((a, b))

    return a

def get_estimated_lmax(num_groups=2, dist=100, group_size=10, scale_factor=4, **kwargs):
    clust_radius = math.sqrt(dist / (num_groups * 2 - 1))
    clust_point_density = (group_size * .50) / (math.pi * clust_radius * clust_radius)
    lmax = scale_factor / math.sqrt(clust_point_density)
    return lmax

def get_triangles_from_list(triangles:np.ndarray, points):
    triangle_list = []
    skip = 3 if triangles.ndim == 1 else 1
    for i in range(0, len(triangles), skip):
        # print(triangles)
        triangle = []
        if triangles.ndim == 1:
            p0 = triangles[i]
            p1 = triangles[i + 1]
            p2 = triangles[i + 2]
        else:
            p0 = triangles[i, 0]
            p1 = triangles[i, 1]
            p2 = triangles[i, 2]
        triangle.append(get_point(p0, points))
        triangle.append(get_point(p1, points))
        triangle.append(get_point(p2, points))
        triangle_list.append(triangle)

    return triangle_list

def get_colored_planar_segments(planes, triangles:np.ndarray, points):
    triangle_meshes = []
    for i,plane in enumerate(planes):
        # Get color for plane mesh
        j = i
        if j >= len(COLOR_PALETTE):
            j = j % len(COLOR_PALETTE)

        plane_mesh = {'color': COLOR_PALETTE[j] , 'triangles': []}
        triangle_list = []
        # Get triangle coordinates
        for t in plane:
            # print(triangles)
            triangle = []
            if triangles.ndim == 1:
                p0 = triangles[t * 3]
                p1 = triangles[t* 3 + 1]
                p2 = triangles[t * 3 + 2]
            else:
                p0 = triangles[t, 0]
                p1 = triangles[t, 1]
                p2 = triangles[t, 2]
            triangle.append(get_point(p0, points))
            triangle.append(get_point(p1, points))
            triangle.append(get_point(p2, points))
            triangle_list.append(triangle)
        plane_mesh['triangles'] = triangle_list
        triangle_meshes.append(plane_mesh)
    return triangle_meshes

def get_all_triangles(delaunay, points):
    triangles = delaunay.triangles
    triangle_coords = get_triangles_from_list(triangles, points)
    return triangle_coords


def scale_points(points, z_value=None, z_scale=1.0):
    if z_value is not None:
        return points[:,0], points[:, 1], np.ones_like(points[:, 1]) * z_value
    else:
        return points[:,0], points[:, 1], points[:,2] * z_scale

def set_up_axes(fig, ax, x_label='X', y_label='Y', z_label='Z', axis_equal=True):
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_zlabel(z_label)
    if axis_equal:
        set_axes_equal(ax)

def plot_planes_3d(lidar_building, triangles, planes, ax, alpha=1.0, z_value=None, z_scale=1.0, color=None):
    ax_planes = []
    for i, plane in enumerate(planes):
        triangles_plane = triangles[plane]
        color_ = COLOR_PALETTE[i] if color is None else color
        color_ = color_ + (alpha, )
        ax_plane = ax.plot_trisurf(*scale_points(lidar_building, z_value=z_value, z_scale=z_scale),triangles=triangles_plane, color=color_,  edgecolor=(0,0,0,0.3), linewidth=0.5)
        ax_planes.append(ax_plane)
    return ax_planes

def plot_polygons_3d(points, polygons, ax, color=None, linewidth=6):
    for i, polygon in enumerate(polygons):
        shell = points[np.array(polygon.shell), :]
        color_ = COLOR_PALETTE[i] if color is None else color
        ax.plot(shell[:, 0], shell[:,1], shell[:, 2], c=color_, linewidth=linewidth, label="Polygon Outline")
        for hole in polygon.holes:
            hole = points[np.array(hole), :]
            ax.plot(hole[:, 0], hole[:,1], hole[:, 2], c='k', linewidth=linewidth)

def plot_triangle_meshes(plane_triangles, ax):
    for plane in plane_triangles:
        color = plane['color']
        triangle_coords = plane['triangles']
        plot_triangles(triangle_coords, ax, color=color, fill=True)

def plot_triangles(triangle_coords, ax, color=COLOR_PALETTE[0], fill=False, ec='black'):
    # print(triangle_coords)
    triangles_shapely = [Polygon(triangle) for triangle in triangle_coords]
    for shape in triangles_shapely:
        patch = PolygonPatch(shape, color=color, ec=ec, alpha=0.5, fill=fill)
        ax.add_patch(patch)

def plot_points(points, ax):
    ax.scatter(points[:, 0], points[:, 1], c='k')


"""Example Multipolygon Extraction
"""
import time
import math

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


from polylidar import MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (plot_polygons_3d, generate_3d_plane, set_axes_equal, plot_planes_3d,
                                     scale_points, rotation_matrix, apply_rotation)


def extract_polygons(opc, classes, config):
    pl = Polylidar3D(**config['polylidar'])
    ga = GaussianAccumulatorS2Beta(level=config['fastga']['level'])
    ico = IcoCharts(level=config['fastga']['level'])
    # 1. Create mesh
    tri_mesh = create_meshes_cuda(
        opc, **config['mesh']['filter'])

    mesh_o3d = create_open_3d_mesh_from_tri_mesh(tri_mesh)
    classes = classes.reshape((np.asarray(mesh_o3d.vertices).shape[0], 1))
    classes_mat = MatrixUInt8(classes)
    tri_mesh.set_vertex_classes(classes_mat, True)
    # alg_timings.update(timings)
    # 2. Get dominant plane normals
    avg_peaks, _, _, _, alg_timings = extract_all_dominant_plane_normals(
        tri_mesh, ga_=ga, ico_chart_=ico, **config['fastga'])

    # alg_timings.update(timings)
    # 3. Extract Planes and Polygons
    planes, obstacles, timings = extract_planes_and_polygons_from_classified_mesh(tri_mesh, avg_peaks, pl_=pl,
                                                                        filter_polygons=True, optimized=True,
                                                                        postprocess=config['polygon']['postprocess'])

    return mesh_o3d, planes, obstacles, timings

def main():
    np.random.seed(1)
    # generate random plane with hole
    plane = generate_3d_plane(bounds_x=[0, 10, 0.5], bounds_y=[0, 10, 0.5], holes=[
        [[3, 5], [3, 5]]], height_noise=0.02, planar_noise=0.02)
    # Generate top of box (causing the hole that we see)
    box_top = generate_3d_plane(bounds_x=[3, 5, 0.2], bounds_y=[3, 5, 0.2], holes=[
    ], height_noise=0.02, height=2, planar_noise=0.02)
    # Generate side of box (causing the hole that we see)
    box_side = generate_3d_plane(bounds_x=[0, 2, 0.2], bounds_y=[
        0, 2, 0.2], holes=[], height_noise=0.02, planar_noise=0.02)
    rm = rotation_matrix([0, 1, 0], -math.pi / 2.0)
    box_side = apply_rotation(rm, box_side) + [5, 3, 0]
    # All points joined together
    points = np.concatenate((plane, box_side, box_top))

    points_mat = MatrixDouble(points)
    polylidar_kwargs = dict(alpha=0.0, lmax=1.0, min_triangles=20, z_thresh=0.1, norm_thresh_min=0.94)
    polylidar = Polylidar3D(**polylidar_kwargs)

    elev = 15.0
    azim = -35

    # Show Point Cloud
    print("Should see point raw point cloud")
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))
    # plot points
    ax.scatter(*scale_points(points), s=20.0, c=points[:, 2], cmap=plt.cm.plasma)
    set_axes_equal(ax)
    ax.view_init(elev=elev, azim=azim)
    fig.savefig("assets/scratch/Basic25DAlgorithm_pointcloud.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_pointcloud.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()

    # Extracts planes and polygons, time
    t1 = time.time()
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.time()
    print("Polylidar Took {:.2f} milliseconds".format((t2 - t1) * 1000))

    triangles = np.asarray(mesh.triangles)
    all_planes = [np.arange(triangles.shape[0])]

    # Show Triangulation
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))

    plot_planes_3d(points, triangles, all_planes, ax, alpha=0.0, z_value=-4.0)
    plot_planes_3d(points, triangles, all_planes, ax, alpha=0.5)
    # plot points
    ax.scatter(*scale_points(points), s=0.1, c='k')
    set_axes_equal(ax, ignore_z=True)
    ax.set_zlim3d([-4, 6])
    ax.view_init(elev=elev, azim=azim)
    print("Should see triangulation point cloud")
    fig.savefig("assets/scratch/Basic25DAlgorithm_mesh.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_mesh.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()

    print("")
    print("Should see two planes extracted, please rotate.")
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))
    # plot all triangles
    plot_polygons_3d(points, polygons, ax)
    plot_planes_3d(points, triangles, planes, ax)
    # plot points
    ax.scatter(*scale_points(points), c='k', s=0.1)
    set_axes_equal(ax)
    ax.view_init(elev=elev, azim=azim)
    fig.savefig("assets/scratch/Basic25DAlgorithm_polygons.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_polygons.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()
    print("")


if __name__ == "__main__":
    main()



"""Example of extracting MultiPolygons with holes of a 2D point set. Step through algorithm.
"""
import time
import numpy as np
from polylidar import Delaunator, MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (generate_test_points, plot_points, plot_triangles, get_estimated_lmax,
                            plot_triangle_meshes, get_triangles_from_list, get_colored_planar_segments, plot_polygons)
import matplotlib.pyplot as plt

np.random.seed(1)

def get_np_buffer_ptr(a):
    pointer, read_only_flag = a.__array_interface__['data']
    return pointer

def main():

    kwargs = dict(num_groups=2, group_size=1000, dist=100.0, seed=1)
    # generate random normally distributed clusters of points, 200 X 2 numpy array.
    points = generate_test_points(**kwargs)
    lmax = get_estimated_lmax(**kwargs)
    polylidar_kwargs = dict(alpha=0.0, lmax=lmax, min_triangles=5)
    # print(polylidar_kwargs)
    # Convert Points and make Polylidar
    points_mat = MatrixDouble(points)
    polylidar = Polylidar3D(**polylidar_kwargs)
    # Extracts planes and polygons, time
    t1 = time.perf_counter()
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.perf_counter()
    triangles = np.asarray(mesh.triangles)
    # print(triangles, triangles.shape)
    triangles = triangles.flatten()
    # print(triangles, triangles.shape)
    planes_np = np.asarray(planes)
    # print(planes_np)
    print("Took {:.2f} milliseconds".format((t2 - t1) * 1000))

    # Plot Data
    if points.shape[0] < 100000:
        fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
        # plot points
        print("Plot Points")
        ax.scatter(points[:, 0], points[:, 1], c='k', s=20.0)
        fig.savefig("assets/scratch/Basic2DAlgorithm_pointcloud.png", bbox_inches='tight', pad_inches=-0.6)
        plt.show()

        print("Plot Mesh")
        fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
        # plot points
        ax.scatter(points[:, 0], points[:, 1], c='k', s=0.1)
        # plot all triangles
        plot_triangles(get_triangles_from_list(triangles, points), ax)
        fig.savefig("assets/scratch/Basic2DAlgorithm_mesh.png", bbox_inches='tight', pad_inches=-0.6)
        plt.show()

        print("Planes and Polygons")
        fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
        # plot points
        ax.scatter(points[:, 0], points[:, 1], c='k', s=0.1)
        # plot all triangles
        plot_triangles(get_triangles_from_list(triangles, points), ax)
        # plot mesh triangles
        triangle_meshes = get_colored_planar_segments(planes_np, triangles, points)
        plot_triangle_meshes(triangle_meshes, ax)
        # plot polygons
        plot_polygons(polygons, points, ax)

        plt.axis('equal')

        plt.show()

if __name__ == "__main__":
    main()


import time
import math

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib 

matplotlib.rc('xtick', labelsize=14) 
matplotlib.rc('ytick', labelsize=14)
# matplotlib.rc('ztick', labelsize=20)
matplotlib.rc('axes', labelsize=18) 


from polylidar import Delaunator, MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (plot_polygons_3d, generate_3d_plane, set_axes_equal, plot_planes_3d,
                           scale_points, rotation_matrix, apply_rotation)



def set_labels(ax):
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

def triangulate_and_combine(points, points_bad, triangles_bad): 
    points_mat = MatrixDouble(points)
    mesh = Delaunator(points_mat)
    mesh.triangulate()

    triangles = np.asarray(mesh.triangles)
    points_new = np.concatenate([points, points_bad], axis=0)
    triangles_new = np.concatenate([triangles, triangles_bad], axis=0)
    all_planes = [np.arange(triangles_new.shape[0])]

    return mesh, points_new, triangles_new, all_planes

def main():
    np.random.seed(1)
    # generate random plane with hole

    z_lim = [-4, 6]
    elev = 10.0
    azim = -35

    points = generate_3d_plane(bounds_x=[2.0, 6.0, 1.0], bounds_y=[2.0, 6.0, 0.5], holes=[], planar_noise=0.0, height_noise=0.01)

    points_bad = np.array([
        [3.0, 3.5, 0.0],
        [3.5, 3.5, 1.0],
        [4.0, 3.5, 0.0]
    ])
    end_point_idx = points.shape[0]
    triangles_bad = np.array([
        [end_point_idx, end_point_idx + 1, end_point_idx + 2]
    ])

    mesh, points_new, triangles_new, all_planes = triangulate_and_combine(points, points_bad, triangles_bad)

    # Show Triangulation
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                    subplot_kw=dict(projection='3d'))

    plot_planes_3d(points_new, triangles_new, all_planes, ax, alpha=0.5)
    set_axes_equal(ax, ignore_z=True)
    ax.set_zlim3d(z_lim)
    ax.view_init(elev=elev, azim=azim)
    plt.show()


    points = generate_3d_plane(bounds_x=[2.0, 6.0, 1.0], bounds_y=[2.0, 6.0, 0.5], holes=[], planar_noise=0.0, height_noise=0.01)
    points[18, 2] = points[18, 2] + 0.30
    points_bad = np.array([
        [3.0, 3.5, 0.0],
        [3.5, 4.0, 0.20],
        [4.0, 3.5, 0.0]
    ])
    end_point_idx = points.shape[0]
    triangles_bad = np.array([
        [end_point_idx, end_point_idx + 1, end_point_idx + 2]
    ])

    mesh, points_new, triangles_new, all_planes = triangulate_and_combine(points, points_bad, triangles_bad)
    # all_planes.append([13, 14])
    # Show Triangulation
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                    subplot_kw=dict(projection='3d'))

    plot_planes_3d(points_new, triangles_new, all_planes, ax, alpha=0.5)
    # plot_planes_3d(points_new, triangles_new, [[triangles_new.shape[0]-1]], ax, alpha=1.0, color=(1,0,0))
    
    ax.plot_trisurf(*scale_points(points_new, z_value=None, z_scale=1.0),triangles=triangles_new[[triangles_new.shape[0]-1]], color=(1,0,0,1.0),  edgecolor=(0,0,0,0.3), linewidth=0.5, antialiased=True, zorder=1.5)
    ax.plot_trisurf(*scale_points(points_new, z_value=None, z_scale=1.0),triangles=triangles_new[[13,14]], color=(0,1,0,0.5),  edgecolor=(0,0,0,0.3), linewidth=0.5, antialiased=False, zorder=0.5)
    # plot_planes_3d(points_new, triangles_new, [[13,14]], ax, alpha=0.9, color=(0,1,0))
    set_axes_equal(ax, ignore_z=True)
    ax.set_zlim3d(z_lim)
    ax.view_init(elev=elev, azim=azim)
    plt.show()




   

if __name__ == "__main__":
    main()



"""Test polygon extraction from a triangular mesh, requires FastGA to be installed
Requires: Fixtures data
"""
import time
import logging
import warnings
import numpy as np
from scipy.spatial.transform import Rotation as R

warnings.filterwarnings("ignore", message="Optimal rotation is not uniquely or poorly defined ")
np.set_printoptions(precision=4, suppress=True)

from examples.python.util.mesh_util import get_mesh_data_iterator
from polylidar import (Polylidar3D, MatrixDouble, MatrixFloat, MatrixInt,
                       create_tri_mesh_copy, bilateral_filter_normals)

from polylidar.polylidarutil.open3d_util import construct_grid, create_lines, flatten, open_3d_mesh_to_trimesh
from polylidar.polylidarutil.plane_filtering import filter_planes_and_holes

from fastga import GaussianAccumulatorS2, MatX3d, IcoCharts
from fastga.peak_and_cluster import find_peaks_from_ico_charts


import open3d as o3d


def filter_and_create_open3d_polygons(points, polygons, rm=None, line_radius=0.01,
                                      config_pp=dict(filter=dict(hole_area=dict(min=0.1, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.25)),
                                                     positive_buffer=0.02, negative_buffer=0.05, simplify=0.02)):
    " Apply polygon filtering algorithm, returns Open3D Mesh Lines"
    # config_pp = dict(filter=dict(hole_area=dict(min=0.00, max=100.0), hole_vertices=dict(min=3), plane_area=dict(min=0.05)),
    #                  positive_buffer=0.00, negative_buffer=0.0, simplify=0.01)
    t1 = time.perf_counter()
    planes, obstacles = filter_planes_and_holes(polygons, points, config_pp, rm=rm)
    t2 = time.perf_counter()
    logging.info("Plane Filtering Took (ms): %.2f", (t2 - t1) * 1000)
    all_poly_lines = create_lines(planes, obstacles, line_radius=line_radius)
    return all_poly_lines, (t2 - t1) * 1000


def down_sample_normals(triangle_normals, down_sample_fraction=0.12, min_samples=10000, flip_normals=False, **kwargs):
    """Return dowsampled normals

    Args:
        triangle_normals (np.ndarray): Triangle Normals
        down_sample_fraction (float, optional): Fraction to downsample. Defaults to 0.12.
        min_samples (int, optional): Minimum number of samples. Defaults to 10000.
        flip_normals (bool, optional): Reverse the normals?. Defaults to False.

    Returns:
        np.ndarray: NX3 downsampled normals
    """
    num_normals = triangle_normals.shape[0]
    to_sample = int(down_sample_fraction * num_normals)
    to_sample = max(min([num_normals, min_samples]), to_sample)
    ds_step = int(num_normals / to_sample)
    triangle_normals_ds = np.ascontiguousarray(triangle_normals[:num_normals:ds_step, :])
    if flip_normals:
        triangle_normals_ds = triangle_normals_ds * -1.0
    return triangle_normals_ds


def get_image_peaks(ico_chart, ga, level=2, with_o3d=False,
                    find_peaks_kwargs=dict(threshold_abs=30, min_distance=1, exclude_border=True, indices=False),
                    cluster_kwargs=dict(t=0.10, criterion='distance'),
                    average_filter=dict(min_total_weight=0.01),
                    **kwargs):

    normalized_bucket_counts_by_vertex = ga.get_normalized_bucket_counts_by_vertex(True)

    t1 = time.perf_counter()
    ico_chart.fill_image(normalized_bucket_counts_by_vertex)  # this takes microseconds
    # plt.imshow(np.asarray(ico_chart.image))
    # plt.show()
    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart, np.asarray(
        normalized_bucket_counts_by_vertex), find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs, average_filter=average_filter)
    t2 = time.perf_counter()

    gaussian_normals_sorted = np.asarray(ico_chart.sphere_mesh.vertices)
    # Create Open3D structures for visualization
    if with_o3d:
        pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
        arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    else:
        pcd_all_peaks = None
        arrow_avg_peaks = None

    elapsed_time = (t2 - t1) * 1000
    timings = dict(t_fastga_peak=elapsed_time)

    logging.debug("Peak Detection - Took (ms): %.2f", (t2 - t1) * 1000)

    return avg_peaks, pcd_all_peaks, arrow_avg_peaks, timings


def extract_all_dominant_planes(tri_mesh, vertices, polylidar_kwargs, config_pp, ds=50, min_samples=10000):
    ga = GaussianAccumulatorS2(level=4, max_phi=180)
    ico = IcoCharts(level=4)

    triangle_normals = np.asarray(tri_mesh.triangle_normals)
    num_normals = triangle_normals.shape[0]
    triangle_normals_ds = down_sample_normals(triangle_normals)

    # Get the data
    t0 = time.perf_counter()
    ga.integrate(MatX3d(triangle_normals_ds))
    t1 = time.perf_counter()
    avg_peaks, _, _, timings = get_image_peaks(ico, ga, level=4, with_o3d=False)
    timings['t_fastga_integrate'] = (t1 - t0) * 1000
    timings['t_fastga_total'] = timings['t_fastga_integrate'] + timings['t_fastga_peak']

    logging.info("Processing mesh with %d triangles", num_normals)
    logging.info("Dominant Plane Normals")
    print(avg_peaks)

    avg_peaks_selected = np.copy(avg_peaks[[0, 1, 2, 3], :])
    pl = Polylidar3D(**polylidar_kwargs)
    avg_peaks_mat = MatrixDouble(avg_peaks_selected)

    # debugging purposes, ignore
    tri_set = pl.extract_tri_set(tri_mesh, avg_peaks_mat)

    t0 = time.perf_counter()
    all_planes, all_polygons = pl.extract_planes_and_polygons_optimized(tri_mesh, avg_peaks_mat)
    t1 = time.perf_counter()
    timings['t_polylidar_planepoly'] = (t1 - t0) * 1000

    all_poly_lines = []
    for i in range(avg_peaks_selected.shape[0]):
        avg_peak = avg_peaks[i, :]
        rm, _ = R.align_vectors([[0, 0, 1]], [avg_peak])
        polygons_for_normal = all_polygons[i]
        # print(polygons_for_normal)
        if len(polygons_for_normal) > 0:
            poly_lines, _ = filter_and_create_open3d_polygons(vertices, polygons_for_normal, rm=rm, config_pp=config_pp)
            all_poly_lines.extend(poly_lines)

    return all_planes, tri_set, all_poly_lines, timings


def run_test(mesh, callback=None, polylidar_kwargs=None, config_pp=None):
    tri_mesh = open_3d_mesh_to_trimesh(mesh)
    bilateral_filter_normals(tri_mesh, 3, 0.1, 0.1)
    vertices = np.asarray(tri_mesh.vertices)
    normals_smooth = np.asarray(tri_mesh.triangle_normals)
    mesh.triangle_normals = o3d.utility.Vector3dVector(normals_smooth)

    planes, tri_set, all_poly_lines, timings = extract_all_dominant_planes(
        tri_mesh, vertices, polylidar_kwargs, config_pp)
    time_polylidar3D = timings['t_polylidar_planepoly']
    polylidar_3d_alg_name = 'Polylidar3D with Provided Mesh'

    mesh_3d_polylidar = []
    mesh_3d_polylidar.extend(flatten([line_mesh.cylinder_segments for line_mesh in all_poly_lines]))
    mesh_3d_polylidar.append(mesh)
    callback(polylidar_3d_alg_name, time_polylidar3D, mesh_3d_polylidar)


def callback(alg_name, execution_time, mesh=None):
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -0.7])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-0.8, translate=[0, 1.0, 0.0])
    logging.info("%s took (ms): %.2f", alg_name, execution_time)
    if mesh:
        if isinstance(mesh, list):
            o3d.visualization.draw_geometries(
                [*mesh, axis_frame])
        else:
            o3d.visualization.draw_geometries([mesh, axis_frame])


def main():
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -1.0])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-1.0, translate=[0, 0.0, 0.0])

    polylidar_kwargs_basement = dict(alpha=0.0, lmax=0.10, min_triangles=80,
                                     z_thresh=0.08, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=6)

    config_pp_basement = dict(filter=dict(hole_area=dict(min=0.05, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.25)),
                              positive_buffer=0.0, negative_buffer=0.025, simplify=0.01)

    polylidar_kwargs_mainfloor = dict(alpha=0.0, lmax=0.10, min_triangles=1000,
                                      z_thresh=0.08, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=6)

    config_pp_main_floor = dict(filter=dict(hole_area=dict(min=0.1, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.25)),
                                positive_buffer=0.02, negative_buffer=0.05, simplify=0.05)

    polylidar_kwargs = [polylidar_kwargs_mainfloor, None, polylidar_kwargs_basement]
    config_pp = [config_pp_main_floor, None, config_pp_basement]

    for i, mesh in enumerate(get_mesh_data_iterator()):
        if i == 1:
            continue
        run_test(mesh, callback=callback, polylidar_kwargs=polylidar_kwargs[i], config_pp=config_pp[i])


if __name__ == "__main__":
    main()



import time
import logging
import warnings
import numpy as np
from copy import deepcopy
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

warnings.filterwarnings("ignore", message="Optimal rotation is not uniquely or poorly defined ")
np.set_printoptions(precision=4, suppress=True)

from examples.python.util.realsense_util import (get_realsense_data, get_frame_data, R_Standard_d400, prep_mesh,
                                                 create_open3d_pc, extract_mesh_planes, COLOR_PALETTE, create_open_3d_mesh)
from examples.python.util.mesh_util import get_mesh_data_iterator

from polylidar import (Polylidar3D, MatrixDouble, MatrixFloat, MatrixInt,
                       create_tri_mesh_copy, bilateral_filter_normals)

from polylidar.polylidarutil.open3d_util import construct_grid, create_lines, flatten, open_3d_mesh_to_trimesh
from polylidar.polylidarutil.plane_filtering import filter_planes_and_holes

from fastga import GaussianAccumulatorS2, MatX3d, IcoCharts
from fastga.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts


import open3d as o3d


def filter_and_create_open3d_polygons(points, polygons, rm=None, line_radius=0.01,
                                      config_pp=dict(filter=dict(hole_area=dict(min=0.1, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.25)),
                                                     positive_buffer=0.02, negative_buffer=0.05, simplify=0.02)):
    " Apply polygon filtering algorithm, return Open3D Mesh Lines "
    # config_pp = dict(filter=dict(hole_area=dict(min=0.00, max=100.0), hole_vertices=dict(min=3), plane_area=dict(min=0.05)),
    #                  positive_buffer=0.00, negative_buffer=0.0, simplify=0.01)
    t1 = time.perf_counter()
    planes, obstacles = filter_planes_and_holes(polygons, points, config_pp, rm=rm)
    t2 = time.perf_counter()
    logging.info("Plane Filtering Took (ms): %.2f", (t2 - t1) * 1000)
    all_poly_lines = create_lines(planes, obstacles, line_radius=line_radius)
    return all_poly_lines, (t2 - t1) * 1000


def down_sample_normals(triangle_normals, down_sample_fraction=0.12, min_samples=10000, flip_normals=False, **kwargs):
    num_normals = triangle_normals.shape[0]
    to_sample = int(down_sample_fraction * num_normals)
    to_sample = max(min([num_normals, min_samples]), to_sample)
    ds_step = int(num_normals / to_sample)
    triangle_normals_ds = np.ascontiguousarray(triangle_normals[:num_normals:ds_step, :])
    if flip_normals:
        triangle_normals_ds = triangle_normals_ds * -1.0
    return triangle_normals_ds

def get_image_peaks(ico_chart, ga, level=2, with_o3d=False,
                    find_peaks_kwargs=dict(threshold_abs=15, min_distance=1, exclude_border=True, indices=False),
                    cluster_kwargs=dict(t=0.10, criterion='distance'),
                    average_filter=dict(min_total_weight=0.01),
                    **kwargs):

    normalized_bucket_counts_by_vertex = ga.get_normalized_bucket_counts_by_vertex(True)

    t1 = time.perf_counter()
    ico_chart.fill_image(normalized_bucket_counts_by_vertex)  # this takes microseconds
    # plt.imshow(np.asarray(ico_chart.image))
    # plt.show()
    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart, np.asarray(
        normalized_bucket_counts_by_vertex), find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs, average_filter=average_filter)
    t2 = time.perf_counter()

    gaussian_normals_sorted = np.asarray(ico_chart.sphere_mesh.vertices)
    # Create Open3D structures for visualization
    if with_o3d:
        pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
        arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    else:
        pcd_all_peaks = None
        arrow_avg_peaks = None

    elapsed_time = (t2 - t1) * 1000
    timings = dict(t_fastga_peak=elapsed_time)

    logging.debug("Peak Detection - Took (ms): %.2f", (t2 - t1) * 1000)


    return avg_peaks, pcd_all_peaks, arrow_avg_peaks, timings


def extract_all_dominant_planes(tri_mesh, vertices, polylidar_kwargs, config_pp, ds=50, min_samples=10000):
    ga = GaussianAccumulatorS2(level=4, max_phi=180)
    ico = IcoCharts(level=4)

    triangle_normals = np.asarray(tri_mesh.triangle_normals)
    num_normals = triangle_normals.shape[0]
    triangle_normals_ds = down_sample_normals(triangle_normals)


    # Get the data
    t0 = time.perf_counter()
    ga.integrate(MatX3d(triangle_normals_ds))
    t1 = time.perf_counter()
    avg_peaks, _, _, timings = get_image_peaks(ico, ga, level=4, with_o3d=False)
    timings['t_fastga_integrate'] = (t1-t0) * 1000
    timings['t_fastga_total'] = timings['t_fastga_integrate'] + timings['t_fastga_peak']


    logging.info("Processing mesh with %d triangles", num_normals)
    logging.info("Dominant Plane Normals")
    print(avg_peaks)

    avg_peaks_selected = np.copy(avg_peaks[[0, 1, 2, 3], :])
    pl = Polylidar3D(**polylidar_kwargs)
    avg_peaks_mat = MatrixDouble(avg_peaks_selected)

    # debugging purposes, ignore
    tri_set = pl.extract_tri_set(tri_mesh, avg_peaks_mat)

    t0 = time.perf_counter()
    all_planes, all_polygons = pl.extract_planes_and_polygons_optimized(tri_mesh, avg_peaks_mat)
    t1 = time.perf_counter()
    timings['t_polylidar_planepoly'] = (t1 - t0) * 1000

    all_poly_lines = []
    for i in range(avg_peaks_selected.shape[0]):
        avg_peak = avg_peaks[i, :]
        rm, _ = R.align_vectors([[0, 0, 1]], [avg_peak])
        polygons_for_normal = all_polygons[i]
        # print(polygons_for_normal)
        if len(polygons_for_normal) > 0:
            poly_lines, _ = filter_and_create_open3d_polygons(vertices, polygons_for_normal, rm=rm, config_pp=config_pp)
            all_poly_lines.extend(poly_lines)

    return all_planes, tri_set, all_poly_lines, timings

def run_test(mesh, callback=None, polylidar_kwargs=None, config_pp=None):
    tri_mesh = open_3d_mesh_to_trimesh(mesh)
    # bilateral_filter_normals(tri_mesh, 3, 0.1, 0.1)
    vertices = np.asarray(tri_mesh.vertices)
    normals_smooth = np.asarray(tri_mesh.triangle_normals)
    mesh.triangle_normals = o3d.utility.Vector3dVector(normals_smooth)

    planes, tri_set, all_poly_lines, timings = extract_all_dominant_planes(
        tri_mesh, vertices, polylidar_kwargs, config_pp)
    time_polylidar3D = timings['t_polylidar_planepoly']
    polylidar_3d_alg_name = 'Polylidar3D with Provided Mesh'

    mesh_3d_polylidar = []
    mesh_3d_polylidar.extend(flatten([line_mesh.cylinder_segments for line_mesh in all_poly_lines]))
    mesh_3d_polylidar.append(mesh)
    callback(polylidar_3d_alg_name, time_polylidar3D, mesh_3d_polylidar)


def callback(alg_name, execution_time, mesh=None):
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -0.7])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-0.8, translate=[0, 1.0, 0.0])
    logging.info("%s took (ms): %.2f", alg_name, execution_time)
    if mesh:
        if isinstance(mesh, list):
            o3d.visualization.draw_geometries(
                [*mesh], width=800, height=500)
        else:
            o3d.visualization.draw_geometries([mesh], width=800, height=500)


def main():
    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.2)
    axis_frame.translate([0, 0.8, -1.0])
    grid_ls = construct_grid(size=2, n=20, plane_offset=-1.0, translate=[0, 0.0, 0.0])

    polylidar_kwargs_basement = dict(alpha=0.0, lmax=0.10, min_triangles=80,
                                     z_thresh=0.08, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=6)

    config_pp_basement = dict(filter=dict(hole_area=dict(min=0.05, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.06)),
                              positive_buffer=0.0, negative_buffer=0.025, simplify=0.01)

    polylidar_kwargs_mainfloor = dict(alpha=0.0, lmax=0.10, min_triangles=1000,
                                      z_thresh=0.08, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=6)

    config_pp_main_floor = dict(filter=dict(hole_area=dict(min=0.1, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.25)),
                                positive_buffer=0.02, negative_buffer=0.05, simplify=0.02)

    polylidar_kwargs = [polylidar_kwargs_mainfloor, None, polylidar_kwargs_basement]
    config_pp = [config_pp_main_floor, None, config_pp_basement]

    for i, mesh in enumerate(get_mesh_data_iterator()):
        if i == 1:
            continue
        run_test(mesh, callback=callback, polylidar_kwargs=polylidar_kwargs[i], config_pp=config_pp[i])


if __name__ == "__main__":
    main()


"""
MainFloor View 1
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 4.6879752961101335, 4.3041731602087356, 1.1239513038042122 ],
			"boundingbox_min" : [ -1.7143359241036764, -4.4827524540588648, -1.2944623277755183 ],
			"field_of_view" : 60.0,
			"front" : [ 0.53039465128384522, -0.61345881428120275, 0.58510665444018772 ],
			"lookat" : [ 1.1083255653557145, 0.45595517159720361, -1.4139722312039875 ],
			"up" : [ -0.37208261429501627, 0.45169957452210624, 0.81087731656270567 ],
			"zoom" : 0.25999999999999956
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}


MainFloor View 2
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 4.6879752961101335, 4.3041731602087356, 1.1239513038042122 ],
			"boundingbox_min" : [ -1.7143359241036764, -4.4827524540588648, -1.2944623277755183 ],
			"field_of_view" : 60.0,
			"front" : [ 0.33172820056487995, -0.73579018685271158, 0.59039749481280723 ],
			"lookat" : [ 0.01577387102866264, 1.2098965809101954, -0.32128930881301754 ],
			"up" : [ -0.24179049551880405, 0.53861595454388955, 0.80711226591320351 ],
			"zoom" : 0.099999999999999617
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}



Basement View 1
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 1.8764505760969685, 3.0280973667097442, 3.045776668203259 ],
			"boundingbox_min" : [ -2.2365574934452548, -3.6804227036671078, -0.71199999999999997 ],
			"field_of_view" : 60.0,
			"front" : [ -0.70399218989352985, 0.54906926280507251, 0.45046413976209737 ],
			"lookat" : [ 0.2914599315270186, -0.61171091088701091, 1.5318870940033267 ],
			"up" : [ 0.33399643257519951, -0.30379868815633204, 0.89227391539903933 ],
			"zoom" : 0.16
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}

Basement View 2
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 1.8764505760969685, 3.0280973667097442, 3.045776668203259 ],
			"boundingbox_min" : [ -2.2365574934452548, -3.6804227036671078, -0.71199999999999997 ],
			"field_of_view" : 60.0,
			"front" : [ 0.81538935556246117, 0.54274210251126309, 0.20142296045160274 ],
			"lookat" : [ -0.94542479921885303, -1.0842317654700471, 1.650916470745605 ],
			"up" : [ -0.17901684954803626, -0.094491727278059712, 0.97929785104119604 ],
			"zoom" : 0.14000000000000001
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}
"""



"""Example of extracting MultiPolygons with holes of a 2D point set
"""
import time

import numpy as np
from polylidar import MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (generate_test_points, plot_triangles, get_estimated_lmax,
                                     plot_triangle_meshes, get_triangles_from_list, get_colored_planar_segments, plot_polygons)
import matplotlib.pyplot as plt

np.random.seed(1)


def main():
    kwargs = dict(num_groups=2, group_size=1000, dist=100.0, seed=1)
    # generate 2 random normally distributed clusters of points, 200 X 2 numpy array.
    points = generate_test_points(**kwargs)
    lmax = get_estimated_lmax(**kwargs)
    polylidar_kwargs = dict(alpha=0.0, lmax=lmax, min_triangles=5)

    # Convert points to matrix format (no copy) and make Polylidar3D Object
    points_mat = MatrixDouble(points, copy=False)
    polylidar = Polylidar3D(**polylidar_kwargs)

    # Extract the mesh, planes, polygons, and time
    t1 = time.perf_counter()
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.perf_counter()

    print("Took {:.2f} milliseconds".format((t2 - t1) * 1000))

    # Convert to numpy format, no copy with np.asarray()
    triangles = np.asarray(mesh.triangles)

    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
    # plot points
    ax.scatter(points[:, 0], points[:, 1], c='k')
    # plot all triangles
    # plt.triplot(points[:,0], points[:,1], triangles) # better alternative
    plot_triangles(get_triangles_from_list(triangles, points), ax)
    # plot seperated planar triangular segments
    triangle_meshes = get_colored_planar_segments(planes, triangles, points)
    plot_triangle_meshes(triangle_meshes, ax)
    # plot polygons
    plot_polygons(polygons, points, ax)
    plt.axis('equal')
    plt.show()


if __name__ == "__main__":
    main()



"""Demonstrates Polylidar3D using Robust Geometric Predicates with Delaunator
You need to have built Polylidar3D with this enabled to see correct polygon generation
"""
import time
import numpy as np
import matplotlib.pyplot as plt
from tests.python.helpers.utils import load_csv
from polylidar import Polylidar3D, MatrixDouble
from polylidar.polylidarutil import (plot_points, plot_triangles,
                                     plot_triangle_meshes, get_triangles_from_list, get_colored_planar_segments, plot_polygons)

# Load point set that for which delaunator generates invalid convex hulls when using non-robust predicates
# Convex hull should be malformed if polylidar not built with robust predicates
def main():
    points = load_csv('robust_1.csv')
    polylidar_kwargs = dict(alpha=0.0, lmax=1000.0, min_triangles=1)

    points_mat = MatrixDouble(points)
    polylidar = Polylidar3D(**polylidar_kwargs)

    # Extracts planes and polygons, time
    t1 = time.time()
    # pick large lmax to get the convex hull, no triangles filtered
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.time()
    print("Took {:.2f} milliseconds".format((t2 - t1) * 1000))
    print("If Robust Geometric Predicates is NOT activated, you should see a malformed convex hull")
    print("See README.md to activate if desired. ~20% performance penalty when active.")
    print("")
    # Convert to numpy format, no copy with np.asarray()
    triangles = np.asarray(mesh.triangles)

    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
    # plot points
    ax.scatter(points[:, 0], points[:, 1], c='k')
    # plot all triangles
    plot_triangles(get_triangles_from_list(triangles, points), ax)
    # plot seperated planar triangular segments
    triangle_meshes = get_colored_planar_segments(planes, triangles, points)
    plot_triangle_meshes(triangle_meshes, ax)
    # plot polygons
    plot_polygons(polygons, points, ax)
    plt.axis('equal')
    plt.show()


if __name__ == "__main__":
    main()



import time
import math
import sys
from os import path, listdir
from os.path import exists, isfile, join, splitext
import re
import logging

import numpy as np
from polylidar.polylidarutil import COLOR_PALETTE
import open3d as o3d

DIR_NAME = path.dirname(__file__)
FIXTURES_DIR = path.join(DIR_NAME, '../../..', 'fixtures')
REALSENSE_DIR = path.join(FIXTURES_DIR, 'realsense')
REALSENSE_COLOR_DIR = path.join(REALSENSE_DIR, 'color')
REALSENSE_DEPTH_DIR = path.join(REALSENSE_DIR, 'depth')
REALSENSE_SCENE_DIR = path.join(REALSENSE_DIR, 'scene')
REALSENSE_TRAJECTORY = path.join(REALSENSE_SCENE_DIR, 'trajectory.log')
REALSENSE_INTRINSICS = path.join(
    REALSENSE_DIR, 'camera_intrinsic_rgb_424.json')


# Conversion between D400 to T265
H_t265_d400 = np.array([
    [1, 0, 0, 0],
    [0, -1.0, 0, 0],
    [0, 0, -1.0, 0],
    [0, 0, 0, 1]])

# Conversion from D400 frame to "Standard" frame (z-axis= to [0,0,1])
R_Standard_d400 = np.array([
    [1, 0, 0, 0],
    [0, 0.0, 1.0, 0],
    [0, -1.0, 0, 0],
    [0, 0, 0, 1]])


logging.basicConfig(level=logging.INFO)


### Read Saved Data from Realsense Directory ###
# Some of this code is pulled from Open3D to read saved RGBD Files
def sorted_alphanum(file_list_ordered):
    def convert(text): return int(text) if text.isdigit() else text
    def alphanum_key(key): return [convert(c)
                                   for c in re.split('([0-9]+)', key)]
    return sorted(file_list_ordered, key=alphanum_key)


def get_file_list(path_, extension=None):
    if extension is None:
        file_list = [path.join(path_, f)
                     for f in listdir(path_) if isfile(join(path_, f))]
    else:
        file_list = [
            path.join(path_, f)
            for f in listdir(path_)
            if isfile(join(path_, f)) and splitext(f)[1] == extension
        ]
    file_list = sorted_alphanum(file_list)
    file_list = [path.abspath(file_) for file_ in file_list]
    return file_list


def get_rgbd_file_lists(path_color=REALSENSE_COLOR_DIR, path_depth=REALSENSE_DEPTH_DIR):
    color_files = get_file_list(path_color, ".jpg") + \
        get_file_list(path_color, ".png")
    depth_files = get_file_list(path_depth, ".png")
    return color_files, depth_files


def read_trajectory(filename):
    traj = []
    with open(filename, 'r') as f:
        metastr = f.readline()
        while metastr:
            mat = np.zeros(shape=(4, 4))
            for i in range(4):
                matstr = f.readline()
                mat[i, :] = np.fromstring(matstr, dtype=float, sep=' \t')
            traj.append(mat)
            metastr = f.readline()
    return traj


def convert_trajectory(traj, inv=True):
    new_traj = []
    for i in range(len(traj)):
        extrinsic_1 = np.linalg.inv(H_t265_d400) @ traj[i] @ H_t265_d400
        if inv:
            extrinsic_1 = np.linalg.inv(extrinsic_1)
        new_traj.append(extrinsic_1)
    return new_traj


def get_realsense_data(color_dir=REALSENSE_COLOR_DIR, depth_dir=REALSENSE_DEPTH_DIR,
                       trajectory_fpath=REALSENSE_TRAJECTORY, intrinsic_fpath=REALSENSE_INTRINSICS):
    color_files, depth_files = get_rgbd_file_lists(
        path_color=color_dir, path_depth=depth_dir)
    traj = convert_trajectory(read_trajectory(trajectory_fpath))
    intrinsics = o3d.io.read_pinhole_camera_intrinsic(intrinsic_fpath)
    return color_files, depth_files, traj, intrinsics

### End Read Saved Data from Realsense Directory ###


def create_open3d_pc(points):
    """ Creates an Open3D point cloud """
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    return pcd


def create_open_3d_mesh(triangles, points, triangle_normals=None, color=COLOR_PALETTE[0], counter_clock_wise=True):
    """Create an Open3D Mesh given triangles vertices

    Arguments:
        triangles {ndarray} -- Triangles array
        points {ndarray} -- Points array

    Keyword Arguments:
        color {list} -- RGB COlor (default: {[1, 0, 0]})

    Returns:
        mesh -- Open3D Mesh
    """
    mesh_2d = o3d.geometry.TriangleMesh()
    if points.ndim == 1:
        points = points.reshape((int(points.shape[0] / 3), 3))
    if triangles.ndim == 1:
        triangles = triangles.reshape((int(triangles.shape[0] / 3), 3))
        # Open 3D expects triangles to be counter clockwise
    if not counter_clock_wise:
        triangles = np.ascontiguousarray(np.flip(triangles, 1))
    mesh_2d.triangles = o3d.utility.Vector3iVector(triangles)
    mesh_2d.vertices = o3d.utility.Vector3dVector(points)
    if triangle_normals is None:
        mesh_2d.compute_vertex_normals()
        mesh_2d.compute_triangle_normals()
    elif triangle_normals.ndim == 1:
        triangle_normals_ = triangle_normals.reshape((int(triangle_normals.shape[0] / 3), 3))
        mesh_2d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals_)
    mesh_2d.paint_uniform_color(color)
    return mesh_2d


def extract_mesh_planes(points, triangles, planes, counter_clock_wise=True, color=None):
    " Converts Polylidar Mesh Planes into Open3D format "

    meshes = []
    color_ = color
    for i, plane in enumerate(planes):
        if color is None:
            color_ = COLOR_PALETTE[i % (len(COLOR_PALETTE) - 1)]
        else:
            color_ = COLOR_PALETTE[0]
        # tris = np.ascontiguousarray(np.flip(triangles[plane, :], 1))
        tris = np.ascontiguousarray(triangles[plane, :])
        mesh = create_open_3d_mesh(tris, points, color=color_, counter_clock_wise=counter_clock_wise)
        meshes.append(mesh)
    return meshes


def get_frame_data(idx, color_files, depth_files, traj, intrinsic, depth_trunc=3.0, stride=2):
    """Gets Frame Data

    Arguments:
        idx {int} -- Index of frame
        color_files {list} -- list of color images
        depth_files {list} -- list of depth images
        traj {list} -- list of extrinsic matrices corresponding to frames
        intrinsic {Open3D intrisics} -- Open3D intrinsics array

    Keyword Arguments:
        depth_trunc {float} -- How much to truncate depth image in meters (default: {3.0})
        stride {int} -- stride for downsampling pont cloud (default: {2})

    Returns:
        tuple -- PointCloud, RGBD Image, extrinsics (D4XX Frame -> Global Standard Frame)
    """
    depth_1 = o3d.io.read_image(depth_files[idx])
    color_1 = o3d.io.read_image(color_files[idx])

    extrinsic = traj[idx]
    rgbd_image_1 = o3d.geometry.RGBDImage.create_from_color_and_depth(
        color_1, depth_1, convert_rgb_to_intensity=False, depth_trunc=depth_trunc)

    pcd_1 = o3d.geometry.PointCloud.create_from_depth_image(
        depth_1, intrinsic, extrinsic, stride=stride, depth_trunc=depth_trunc)
    return pcd_1, rgbd_image_1, R_Standard_d400 @ np.linalg.inv(extrinsic)


def prep_mesh(mesh):
    " Prepares mesh for visualization "
    mesh_list = mesh
    if not isinstance(mesh_list, list):
        mesh_list = [mesh]
    for mesh_ in mesh_list:
        mesh_.compute_triangle_normals()
        mesh_.paint_uniform_color(COLOR_PALETTE[0])



import json 
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

sns.set(font_scale=1.15)
sns.set_style('white')
sns.set_style("ticks")



def plot_df(df, name='basement'):
    fig = plt.figure(figsize=(6,3.5))
    ax = sns.lineplot(x="threads", y="Speedup", hue="normals", palette='deep', data=df, legend=False)
    ax2 = ax.twinx()
    sns.lineplot(x="threads", y="Execution Time (ms)", ax=ax2, hue='normals',palette='muted', data=df, legend=False)
    for line in ax2.lines:
        line.set_linestyle('--')

    current_palette = sns.color_palette()
    from matplotlib.lines import Line2D
    custom_lines = [Line2D([0], [0], color=current_palette[i], lw=1) for i in range(4)]
    custom_lines_2 = [Line2D([0], [0], color='k', lw=1),
                    Line2D([0], [0], color='k', lw=1, dashes=[2,6])]

    ax.legend(custom_lines, ['1 DP', '2 DP', '3 DP', '4 DP'], loc='upper left', bbox_to_anchor=(0.1, 1,))
    ax2.legend(custom_lines_2, ['Speedup', 'Execution Time'], loc='upper left', bbox_to_anchor=(0.35, 1,))
    # ax.axis('equal')
    # ax.legend(custom_lines, ['1 DP', '2 DP', '3 DP', '4 DP'], loc='upper left', bbox_to_anchor=(0.4, 0.4,))
    # ax2.legend(custom_lines_2, ['Speedup', 'Execution Time'], loc='upper left', bbox_to_anchor=(0.6, 0.4,))
    fig.savefig(f'assets/scratch/{name}_speedup.pdf', bbox_inches='tight')
    plt.tight_layout()
    plt.show()


def main():
    with open('bench/records/bench_mesh_paper.json') as json_file:
        data = json.load(json_file)
    benchmarks = data['benchmarks']

    benchmarks_filtered = []

    for i, bench in enumerate(benchmarks):
        name = bench['name']
        run_name = bench['run_name']
        aggregate_name = bench['aggregate_name']
        if i % 3 == 0:
            names = name.split('/')
            bench['mesh'] = names[0]
            bench['normals'] = int(names[2])
            bench['threads'] = int(names[3])
            bench['time_mean'] = bench['real_time'] if bench['time_unit'] == 'ms' else bench['real_time'] / 1000
            bench_median = benchmarks[i+1]
            bench['time_median'] = bench_median['real_time'] if bench_median['time_unit'] == 'ms' else bench_median['real_time'] / 1000
            bench_std = benchmarks[i+2]
            bench['time_std'] = bench_std['real_time'] if bench_std['time_unit'] == 'ms' else bench_std['real_time'] / 1000
            benchmarks_filtered.append(bench)
    # calculate speedup
    current_speed = None
    for i, bench in enumerate(benchmarks_filtered):
        # reset current speed every 8 iterations (8 threads)
        if i % 8 == 0:
            current_speed = bench['time_mean']
        bench['speedup'] = current_speed / bench['time_mean']

    df = pd.DataFrame.from_records(benchmarks_filtered)
    df_reduced = df[['mesh', 'normals', 'threads', 'speedup', 'time_mean', 'time_median', 'time_std']]
    df_reduced = df_reduced.rename(columns={'time_mean': "Execution Time (ms)", "speedup": "Speedup"})

    df_sparse = df_reduced[df_reduced['mesh'] == 'SparseMeshPaper']
    df_dense = df_reduced[df_reduced['mesh'] == 'DenseMeshPaper']

    plot_df(df_sparse, 'basement')
    plot_df(df_dense, 'mainfloor')

if __name__ == "__main__":
    main()


import time
import pickle
import numpy as np
from polylidar import Delaunator, MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (generate_test_points, plot_points, plot_triangles, get_estimated_lmax,
                                     plot_triangle_meshes, get_triangles_from_list, get_colored_planar_segments, plot_polygons)
import matplotlib.pyplot as plt

np.random.seed(1)


def get_np_buffer_ptr(a):
    pointer, read_only_flag = a.__array_interface__['data']
    return pointer


def stitch_letters(letters_map, letters=['P', 'L']):
    # letters = ['P', 'O', 'L', 'Y', 'L', 'I', 'D', 'A', 'R']
    x_offset = np.array([0.0, 0.0])
    x_spacing = np.array([-30.0, 0.0])
    all_points = []
    for letter in letters:
        letter_points = np.copy(letters_map[letter])
        letter_points = letter_points + x_offset + x_spacing
        x_offset[0] = letter_points.max(axis=0)[0]
        all_points.append(letter_points)

    all_points = np.concatenate(all_points, axis=0)

    return all_points

def read_alphabet(file_name='fixtures/pointsets/alphabet_2000.pkl'):
    with open(file_name, 'rb') as file:
        data = pickle.load(file)
    letters = ['P', 'O', 'L', 'Y', 'I', 'D', 'A', 'R']
    letter_map = dict()
    for letter in letters:
        letter_map[letter] = [item for item in data if item['poly_param']['name'] == letter][0]['points']
    
    return letter_map


def main():

    letter_map = read_alphabet()
    # letters = ['P', 'O', 'L', 'Y', 'L', 'I', 'D', 'A', 'R']
    points = stitch_letters(letter_map)

    # plt.scatter(points[:, 0], points[:, 1], s=0.2)
    # plt.show()
    print(int(points.shape[0] / 2))
    idx = np.random.randint(0, points.shape[0], size=int(points.shape[0] / 2))
    points = np.ascontiguousarray(points[idx, :])
    lmax = 10.0
    polylidar_kwargs = dict(alpha=0.0, lmax=lmax, min_triangles=5)
    # print(polylidar_kwargs)
    # Convert Points and make Polylidar
    points_mat = MatrixDouble(points)
    polylidar = Polylidar3D(**polylidar_kwargs)
    # Extracts planes and polygons, time
    t1 = time.perf_counter()
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.perf_counter()
    triangles = np.asarray(mesh.triangles)
    # print(triangles, triangles.shape)
    triangles = triangles.flatten()
    # print(triangles, triangles.shape)
    planes_np = planes
    # print(planes_np)
    print("Took {:.2f} milliseconds".format((t2 - t1) * 1000))

    # Plot Data
    if points.shape[0] < 100000:
        fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.axis('equal')
        plt.axis('off')
        # plot points
        plot_points(points, ax)
        fig.savefig('assets/scratch/pl_logo_points.png', bbox_inches='tight', transparent=True)
        # plot polygons
        plot_polygons(polygons, points, ax, linewidth=6.0)
        fig.savefig('assets/scratch/pl_logo_poly.png', bbox_inches='tight', transparent=True)
        # plot all triangles
        plot_triangles(get_triangles_from_list(triangles, points), ax)
        # plot mesh triangles
        triangle_meshes = get_colored_planar_segments(planes_np, triangles, points)
        plot_triangle_meshes(triangle_meshes, ax)
        # plot polygons
        plot_polygons(polygons, points, ax, linewidth=4.0)

        plt.subplots_adjust(wspace=0.185, hspace=0.185,left=0.110,top=0.535,right=0.750,bottom=0.110) 
        fig.savefig('assets/scratch/pl_logo.png', bbox_inches='tight', transparent=True)
        plt.show()


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
from tests.python.helpers.utils import load_csv, load_npy


def plot_points(points):

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.scatter(points[:,0], points[:, 1], s=1)
    plt.axis('off')
    ax.axis('equal')

    plt.show()


def main():
    points_1 = load_csv('caholes_8000.csv', delimeter=' ')
    plot_points(points_1)

    points_1 = load_csv('ca_8000.csv', delimeter=' ')
    plot_points(points_1)


if __name__ == "__main__":
    main()


import numpy as np
import open3d as o3d

from .line_mesh import LineMesh
from polylidar.polylidarutil import COLOR_PALETTE
from polylidar import MatrixDouble, MatrixInt, create_tri_mesh_copy
EXTRINSICS = None

MAX_POLYS = 10
ORANGE = (255 / 255, 188 / 255, 0)
GREEN = (0, 255 / 255, 0)


def open_3d_mesh_to_trimesh(mesh: o3d.geometry.TriangleMesh):
    triangles = np.asarray(mesh.triangles)
    vertices = np.asarray(mesh.vertices)

    triangles = np.ascontiguousarray(triangles)

    vertices_mat = MatrixDouble(vertices)
    triangles_mat = MatrixInt(triangles)
    triangles_mat_np = np.asarray(triangles_mat)

    tri_mesh = create_tri_mesh_copy(vertices_mat, triangles_mat)
    return tri_mesh

def create_open_3d_mesh_from_tri_mesh(tri_mesh):
    """Create an Open3D Mesh given a Polylidar TriMesh"""
    triangles = np.asarray(tri_mesh.triangles)
    vertices = np.asarray(tri_mesh.vertices)
    triangle_normals = np.asarray(tri_mesh.triangle_normals)
    return create_open_3d_mesh(triangles, vertices, triangle_normals, counter_clock_wise=tri_mesh.counter_clock_wise)


def create_open_3d_mesh(triangles, points, triangle_normals=None, color=COLOR_PALETTE[0], counter_clock_wise=True):
    """Create an Open3D Mesh given triangles vertices

    Arguments:
        triangles {ndarray} -- Triangles array
        points {ndarray} -- Points array

    Keyword Arguments:
        color {list} -- RGB COlor (default: {[1, 0, 0]})

    Returns:
        mesh -- Open3D Mesh
    """
    mesh_o3d = o3d.geometry.TriangleMesh()
    if points.ndim == 1:
        points = points.reshape((int(points.shape[0] / 3), 3))
    if triangles.ndim == 1:
        triangles = triangles.reshape((int(triangles.shape[0] / 3), 3))
        # Open 3D expects triangles to be counter clockwise
    if not counter_clock_wise:
        triangles = np.ascontiguousarray(np.flip(triangles, 1))
    mesh_o3d.triangles = o3d.utility.Vector3iVector(triangles)

    mask = np.isnan(points).any(axis=1) # I think that we need this with open3d 0.10.0
    points[mask, :] = [0,0,0]
    mesh_o3d.vertices = o3d.utility.Vector3dVector(points)
    if triangle_normals is None:
        mesh_o3d.compute_vertex_normals()
        mesh_o3d.compute_triangle_normals()
    elif triangle_normals.ndim == 1:
        triangle_normals_ = triangle_normals.reshape((int(triangle_normals.shape[0] / 3), 3))
        mesh_o3d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals_)
    else:
        mesh_o3d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals)
    mesh_o3d.paint_uniform_color(color)
    mesh_o3d.compute_vertex_normals()
    return mesh_o3d

def flatten(l): return [item for sublist in l for item in sublist]


def update_points(pcd, pc):
    pcd.points = o3d.utility.Vector3dVector(pc)


def set_line(line_set, points, lines, colors):
    line_set.points = o3d.utility.Vector3dVector(points)
    line_set.lines = o3d.utility.Vector2iVector(lines)
    line_set.colors = o3d.utility.Vector3dVector(colors)


def construct_grid(size=10, n=10, color=[0.5, 0.5, 0.5], plane='xy', plane_offset=-1, translate=[0, 0, 0]):
    grid_ls = o3d.geometry.LineSet()
    my_grid = make_grid(size=size, n=n, color=color, plane=plane, plane_offset=plane_offset, translate=translate)
    set_line(grid_ls, *my_grid)
    return grid_ls


def make_grid(size=10, n=10, color=[0.5, 0.5, 0.5], plane='xy', plane_offset=-1, translate=[0, 0, 0]):
    """draw a grid as a line set"""

    # lineset = o3d.geometry.LineSet()
    s = size / float(n)
    s2 = 0.5 * size
    points = []

    for i in range(0, n + 1):
        x = -s2 + i * s
        points.append([x, -s2, plane_offset])
        points.append([x, s2, plane_offset])
    for i in range(0, n + 1):
        z = -s2 + i * s
        points.append([-s2, z, plane_offset])
        points.append([s2, z, plane_offset])

    points = np.array(points)
    if plane == 'xz':
        points[:, [2, 1]] = points[:, [1, 2]]

    points = points + translate

    n_points = points.shape[0]
    lines = [[i, i + 1] for i in range(0, n_points - 1, 2)]
    colors = [list(color)] * (n_points - 1)
    return points, lines, colors


def clear_polys(all_polys, vis):
    for line_mesh in all_polys:
        line_mesh.remove_line(vis)
    return []


def handle_shapes(vis, planes, obstacles, all_polys, line_radius=0.15):
    all_polys = clear_polys(all_polys, vis)
    for plane, _ in planes:
        points = np.array(plane.exterior)
        line_mesh = LineMesh(points, colors=GREEN, radius=line_radius)
        line_mesh.add_line(vis)
        all_polys.append(line_mesh)

    for plane, _ in obstacles:
        points = np.array(plane.exterior)
        line_mesh = LineMesh(points, colors=ORANGE, radius=line_radius)
        line_mesh.add_line(vis)
        all_polys.append(line_mesh)

    return all_polys


def create_lines(planes, obstacles, line_radius=0.15, rotate_func=None):
    all_polys = []
    for plane, _ in planes:
        points = np.array(plane.exterior)
        if rotate_func:
            points = rotate_func(points)
        line_mesh = LineMesh(points, colors=GREEN, radius=line_radius)
        all_polys.append(line_mesh)

    for plane, _ in obstacles:
        points = np.array(plane.exterior)
        if rotate_func:
            points = rotate_func(points)
        line_mesh = LineMesh(points, colors=ORANGE, radius=line_radius)
        all_polys.append(line_mesh)

    return all_polys


def get_extrinsics(vis):
    ctr = vis.get_view_control()
    camera_params = ctr.convert_to_pinhole_camera_parameters()
    return camera_params.extrinsic


def set_initial_view(vis, extrinsics=[EXTRINSICS]):
    ctr = vis.get_view_control()
    camera_params = ctr.convert_to_pinhole_camera_parameters()
    camera_params.extrinsic = extrinsics
    ctr.convert_from_pinhole_camera_parameters(camera_params)



import time
import logging
from copy import deepcopy

import numpy as np
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

from polylidar.polylidarutil.plane_filtering import filter_planes_and_holes
from polylidar import MatrixDouble, Polylidar3D
from polylidar.polylidarutil.open3d_util import create_lines

from fastga import GaussianAccumulatorS2, MatX3d, IcoCharts
from fastga.peak_and_cluster import find_peaks_from_ico_charts
from fastga.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals


import open3d as o3d


def down_sample_normals(triangle_normals, down_sample_fraction=0.12, min_samples=100, flip_normals=False, **kwargs):
    num_normals = triangle_normals.shape[0]
    to_sample = int(down_sample_fraction * num_normals)
    to_sample = max(min([num_normals, min_samples]), to_sample)
    ds_step = int(num_normals / to_sample)
    triangle_normals_ds = np.ascontiguousarray(triangle_normals[:num_normals:ds_step, :])
    if flip_normals:
        triangle_normals_ds = triangle_normals_ds * -1.0
    return triangle_normals_ds


def get_image_peaks(ico_chart, ga, level=2, with_o3d=False,
                    find_peaks_kwargs=dict(threshold_abs=2, min_distance=1, exclude_border=False, indices=False),
                    cluster_kwargs=dict(t=0.10, criterion='distance'),
                    average_filter=dict(min_total_weight=0.01),
                    **kwargs):

    normalized_bucket_counts_by_vertex = ga.get_normalized_bucket_counts_by_vertex(True)

    t1 = time.perf_counter()
    ico_chart.fill_image(normalized_bucket_counts_by_vertex)  # this takes microseconds
    # plt.imshow(np.asarray(ico_chart.image))
    # plt.show()
    average_vertex_normals = np.asarray(ga.get_average_normals_by_vertex(
        True)) if hasattr(ga, 'get_average_normals_by_vertex') else None
    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart, np.asarray(
        normalized_bucket_counts_by_vertex), average_vertex_normals, find_peaks_kwargs, cluster_kwargs, average_filter)
    t2 = time.perf_counter()

    gaussian_normals_sorted = np.asarray(ico_chart.sphere_mesh.vertices)
    # Create Open3D structures for visualization
    if with_o3d:
        pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
        arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    else:
        pcd_all_peaks = None
        arrow_avg_peaks = None

    elapsed_time = (t2 - t1) * 1000
    timings = dict(t_fastga_peak=elapsed_time)

    logging.debug("Peak Detection - Took (ms): %.2f", (t2 - t1) * 1000)

    return avg_peaks, pcd_all_peaks, arrow_avg_peaks, timings


def extract_all_dominant_plane_normals(tri_mesh, level=5, with_o3d=False, ga_=None, ico_chart_=None, **kwargs):

    # Reuse objects if provided
    if ga_ is not None:
        ga = ga_
    else:
        ga = GaussianAccumulatorS2(level=level)

    if ico_chart_ is not None:
        ico_chart = ico_chart_
    else:
        ico_chart = IcoCharts(level=level)

    triangle_normals = np.asarray(tri_mesh.triangle_normals)
    triangle_normals_ds = down_sample_normals(triangle_normals, **kwargs)

    # np.savetxt('bad_normals.txt', triangle_normals_ds)
    triangle_normals_ds_mat = MatX3d(triangle_normals_ds)
    t1 = time.perf_counter()
    ga.integrate(triangle_normals_ds_mat)
    t2 = time.perf_counter()

    logging.debug("Gaussian Accumulator - Normals Sampled: %d; Took (ms): %.2f",
                  triangle_normals_ds.shape[0], (t2 - t1) * 1000)

    avg_peaks, pcd_all_peaks, arrow_avg_peaks, timings_dict = get_image_peaks(
        ico_chart, ga, level=level, with_o3d=with_o3d, **kwargs)

    # Create Open3D structures for visualization
    if with_o3d:
        # Visualize the Sphere
        accumulator_counts = np.asarray(ga.get_normalized_bucket_counts())
        refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga.mesh.triangles), np.asarray(ga.mesh.vertices))
        color_counts = get_colors(accumulator_counts)[:, :3]
        colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, color_counts)
    else:
        colored_icosahedron = None

    elapsed_time_fastga = (t2 - t1) * 1000
    elapsed_time_peak = timings_dict['t_fastga_peak']
    elapsed_time_total = elapsed_time_fastga + elapsed_time_peak

    timings = dict(t_fastga_total=elapsed_time_total,
                   t_fastga_integrate=elapsed_time_fastga, t_fastga_peak=elapsed_time_peak)

    ga.clear_count()
    return avg_peaks, pcd_all_peaks, arrow_avg_peaks, colored_icosahedron, timings


def filter_and_create_polygons(points, polygons, rm=None, line_radius=0.005,
                               postprocess=dict(filter=dict(hole_area=dict(min=0.025, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.05)),
                                                positive_buffer=0.00, negative_buffer=0.00, simplify=0.0)):
    " Apply polygon filtering algorithm, return Open3D Mesh Lines "
    t1 = time.perf_counter()
    planes, obstacles = filter_planes_and_holes(polygons, points, postprocess, rm=rm)
    t2 = time.perf_counter()
    return planes, obstacles, (t2 - t1) * 1000


def extract_planes_and_polygons_from_mesh(tri_mesh, avg_peaks,
                                          polylidar_kwargs=dict(alpha=0.0, lmax=0.1, min_triangles=2000,
                                                                z_thresh=0.1, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=50, task_threads=4),
                                          filter_polygons=True, pl_=None, optimized=False,
                                          postprocess=dict(filter=dict(hole_area=dict(min=0.025, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.05)),
                                                           positive_buffer=0.00, negative_buffer=0.00, simplify=0.0)):

    if pl_ is not None:
        pl = pl_
    else:
        pl = Polylidar3D(**polylidar_kwargs)

    avg_peaks_mat = MatrixDouble(avg_peaks)
    t0 = time.perf_counter()
    if optimized:
        all_planes, all_polygons = pl.extract_planes_and_polygons_optimized(tri_mesh, avg_peaks_mat)
    else:
        all_planes, all_polygons = pl.extract_planes_and_polygons(tri_mesh, avg_peaks_mat)
    t1 = time.perf_counter()
    # tri_set = pl.extract_tri_set(tri_mesh, avg_peaks_mat)
    # planes_tri_set = [np.argwhere(np.asarray(tri_set) == i)  for i in range(1, 2)]
    # o3d_mesh_painted = paint_planes(o3d_mesh, planes_tri_set)
    polylidar_time = (t1 - t0) * 1000
    # logging.info("Polygon Extraction - Took (ms): %.2f", polylidar_time)
    all_planes_shapely = []
    all_obstacles_shapely = []
    time_filter = []
    # all_poly_lines = []
    if filter_polygons:
        vertices = np.asarray(tri_mesh.vertices)
        for i in range(avg_peaks.shape[0]):
            avg_peak = avg_peaks[i, :]
            rm, _ = R.align_vectors([[0, 0, 1]], [avg_peak])
            polygons_for_normal = all_polygons[i]
            # print(polygons_for_normal)
            if len(polygons_for_normal) > 0:
                planes_shapely, obstacles_shapely, filter_time = filter_and_create_polygons(
                    vertices, polygons_for_normal, rm=rm, postprocess=postprocess)
                all_planes_shapely.extend(planes_shapely)
                all_obstacles_shapely.extend(obstacles_shapely)
                time_filter.append(filter_time)
                # all_poly_lines.extend(poly_lines)
    timings = dict(t_polylidar_planepoly=polylidar_time, t_polylidar_filter=np.array(time_filter).sum())
    # all_planes_shapely, all_obstacles_shapely, all_poly_lines, timings
    return all_planes_shapely, all_obstacles_shapely, timings


def extract_planes_and_polygons_from_classified_mesh(tri_mesh, avg_peaks,
                                                     polylidar_kwargs=dict(alpha=0.0, lmax=0.1, min_triangles=2000,
                                                                           z_thresh=0.1, norm_thresh=0.95, norm_thresh_min=0.95, min_hole_vertices=50, task_threads=4),
                                                     filter_polygons=True, pl_=None, optimized=True,
                                                     postprocess=dict(filter=dict(hole_area=dict(min=0.025, max=100.0), hole_vertices=dict(min=6), plane_area=dict(min=0.05)),
                                                                      positive_buffer=0.00, negative_buffer=0.00, simplify=0.0)):

    if pl_ is not None:
        pl = pl_
    else:
        pl = Polylidar3D(**polylidar_kwargs)

    avg_peaks_mat = MatrixDouble(avg_peaks)
    t0 = time.perf_counter()
    all_planes, all_polygons = pl.extract_planes_and_polygons_optimized_classified(tri_mesh, avg_peaks_mat)
    t1 = time.perf_counter()

    polylidar_time = (t1 - t0) * 1000
    all_planes_shapely = []
    all_obstacles_shapely = []
    time_filter = []
    # all_poly_lines = []
    if filter_polygons:
        vertices = np.asarray(tri_mesh.vertices)
        for i in range(avg_peaks.shape[0]):
            avg_peak = avg_peaks[i, :]
            rm, _ = R.align_vectors([[0, 0, 1]], [avg_peak])
            polygons_for_normal = all_polygons[i]
            # print(polygons_for_normal)
            if len(polygons_for_normal) > 0:
                planes_shapely, obstacles_shapely, filter_time = filter_and_create_polygons(
                    vertices, polygons_for_normal, rm=rm, postprocess=postprocess)
                all_planes_shapely.extend(planes_shapely)
                all_obstacles_shapely.extend(obstacles_shapely)
                time_filter.append(filter_time)
                # all_poly_lines.extend(poly_lines)

    timings = dict(t_polylidar_planepoly=polylidar_time, t_polylidar_filter=np.array(time_filter).sum())
    # all_planes_shapely, all_obstacles_shapely, all_poly_lines, timings
    return all_planes_shapely, all_obstacles_shapely, timings



"""Module which creates mesh lines from a line set
Open3D relies upon using glLineWidth to set line width on a LineSet
However, this method is now deprecated and not fully supporeted in newer OpenGL versions
See:
    Open3D Github Pull Request - https://github.com/intel-isl/Open3D/pull/738
    Other Framework Issues - https://github.com/openframeworks/openFrameworks/issues/3460

This module aims to solve this by converting a line into a triangular mesh (which has thickness)
The basic idea is to create a cylinder for each line segment, translate it, and then rotate it.

License: MIT

"""
import numpy as np
import open3d as o3d
from open3d import __version__ as o3d_version

o3d_major_version = int(o3d_version.split('.')[1])

def align_vector_to_another(a=np.array([0, 0, 1]), b=np.array([1, 0, 0])):
    """
    Aligns vector a to vector b with axis angle rotation
    """
    if np.array_equal(a, b):
        return None, None
    axis_ = np.cross(a, b)
    axis_ = axis_ / np.linalg.norm(axis_)
    angle = np.arccos(np.dot(a, b))

    return axis_, angle


def normalized(a, axis=-1, order=2):
    """Normalizes a numpy array of points"""
    l2 = np.atleast_1d(np.linalg.norm(a, order, axis))
    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis), l2


class LineMesh(object):
    def __init__(self, points, lines=None, colors=[0, 1, 0], radius=0.15):
        """Creates a line represented as sequence of cylinder triangular meshes

        Arguments:
            points {ndarray} -- Numpy array of ponts Nx3.

        Keyword Arguments:
            lines {list[list] or None} -- List of point index pairs denoting line segments. If None, implicit lines from ordered pairwise points. (default: {None})
            colors {list} -- list of colors, or single color of the line (default: {[0, 1, 0]})
            radius {float} -- radius of cylinder (default: {0.15})
        """
        self.points = np.array(points)
        self.lines = np.array(
            lines) if lines is not None else self.lines_from_ordered_points(self.points)
        self.colors = np.array(colors)
        self.radius = radius
        self.cylinder_segments = []

        self.create_line_mesh()

    @staticmethod
    def lines_from_ordered_points(points):
        lines = [[i, i + 1] for i in range(0, points.shape[0] - 1, 1)]
        return np.array(lines)

    def create_line_mesh(self):
        first_points = self.points[self.lines[:, 0], :]
        second_points = self.points[self.lines[:, 1], :]
        line_segments = second_points - first_points
        line_segments_unit, line_lengths = normalized(line_segments)

        z_axis = np.array([0, 0, 1])
        # Create triangular mesh cylinder segments of line
        for i in range(line_segments_unit.shape[0]):
            line_segment = line_segments_unit[i, :]
            line_length = line_lengths[i]
            # get axis angle rotation to allign cylinder with line segment
            axis, angle = align_vector_to_another(z_axis, line_segment)
            # Get translation vector
            translation = first_points[i, :] + line_segment * line_length * 0.5
            # create cylinder and apply transformations
            cylinder_segment = o3d.geometry.TriangleMesh.create_cylinder(
                self.radius, line_length)
            cylinder_segment = cylinder_segment.translate(
                translation, relative=False)
            if axis is not None:
                axis_a = axis * angle
                rotation_3x3 = cylinder_segment.get_rotation_matrix_from_axis_angle(axis_a)
                center = cylinder_segment.get_center() if o3d_major_version > 9 else True
                cylinder_segment = cylinder_segment.rotate(rotation_3x3, center=center)
            # color cylinder
            color = self.colors if self.colors.ndim == 1 else self.colors[i, :]
            cylinder_segment.paint_uniform_color(color)

            self.cylinder_segments.append(cylinder_segment)

    def add_line(self, vis):
        """Adds this line to the visualizer"""
        for cylinder in self.cylinder_segments:
            vis.add_geometry(cylinder)

    def remove_line(self, vis):
        """Removes this line from the visualizer"""
        for cylinder in self.cylinder_segments:
            vis.remove_geometry(cylinder)


def main():
    print("Demonstrating LineMesh vs LineSet")
    # Create Line Set
    points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1],
              [0, 1, 1], [1, 1, 1]]
    lines = [[0, 1], [0, 2], [1, 3], [2, 3], [4, 5], [4, 6], [5, 7], [6, 7],
             [0, 4], [1, 5], [2, 6], [3, 7]]
    colors = [[1, 0, 0] for i in range(len(lines))]

    line_set = o3d.geometry.LineSet()
    line_set.points = o3d.utility.Vector3dVector(points)
    line_set.lines = o3d.utility.Vector2iVector(lines)
    line_set.colors = o3d.utility.Vector3dVector(colors)

    # Create Line Mesh 1
    points = np.array(points) + [0, 0, 2]
    line_mesh1 = LineMesh(points, lines, colors, radius=0.02)
    line_mesh1_geoms = line_mesh1.cylinder_segments

    # Create Line Mesh 2
    points = np.array(points) + [0, 2, 0]
    line_mesh2 = LineMesh(points, radius=0.03)
    line_mesh2_geoms = line_mesh2.cylinder_segments

    o3d.visualization.draw_geometries(
        [line_set, *line_mesh1_geoms, *line_mesh2_geoms])


if __name__ == "__main__":
    main()



"""Example MultiPolygon Extraction with an **Unorganized* 3D Point Cloud. Step through algorithm.
Note this method is only suitable if you know the dominant surface normal of the plane you desire to extract.
The 3D point cloud must be rotated such that this surface normal is [0,0,1], i.e., the plane is aligned with the XY Plane
The method relies upon 3D->2D Projection and performs 2.5D Delaunay Triangulation. 
"""
import time
import math

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


from polylidar import MatrixDouble, Polylidar3D
from polylidar.polylidarutil import (plot_polygons_3d, generate_3d_plane, set_axes_equal, plot_planes_3d,
                                     scale_points, rotation_matrix, apply_rotation)


def main():
    np.random.seed(1)
    # generate random plane with hole
    plane = generate_3d_plane(bounds_x=[0, 10, 0.5], bounds_y=[0, 10, 0.5], holes=[
        [[3, 5], [3, 5]]], height_noise=0.02, planar_noise=0.02)
    # Generate top of box (causing the hole that we see)
    box_top = generate_3d_plane(bounds_x=[3, 5, 0.2], bounds_y=[3, 5, 0.2], holes=[
    ], height_noise=0.02, height=2, planar_noise=0.02)
    # Generate side of box (causing the hole that we see)
    box_side = generate_3d_plane(bounds_x=[0, 2, 0.2], bounds_y=[
        0, 2, 0.2], holes=[], height_noise=0.02, planar_noise=0.02)
    rm = rotation_matrix([0, 1, 0], -math.pi / 2.0)
    box_side = apply_rotation(rm, box_side) + [5, 3, 0]
    # All points joined together
    points = np.concatenate((plane, box_side, box_top))

    points_mat = MatrixDouble(points)
    polylidar_kwargs = dict(alpha=0.0, lmax=1.0, min_triangles=20, z_thresh=0.1, norm_thresh_min=0.94)
    polylidar = Polylidar3D(**polylidar_kwargs)

    elev = 15.0
    azim = -35

    # Show Point Cloud
    print("Should see point raw point cloud")
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))
    # plot points
    ax.scatter(*scale_points(points), s=20.0, c=points[:, 2], cmap=plt.cm.plasma)
    set_axes_equal(ax)
    ax.view_init(elev=elev, azim=azim)
    fig.savefig("assets/scratch/Basic25DAlgorithm_pointcloud.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_pointcloud.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()

    # Extracts planes and polygons, time
    t1 = time.time()
    mesh, planes, polygons = polylidar.extract_planes_and_polygons(points_mat)
    t2 = time.time()
    print("Polylidar Took {:.2f} milliseconds".format((t2 - t1) * 1000))

    triangles = np.asarray(mesh.triangles)
    all_planes = [np.arange(triangles.shape[0])]

    # Show Triangulation
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))

    plot_planes_3d(points, triangles, all_planes, ax, alpha=0.0, z_value=-4.0)
    plot_planes_3d(points, triangles, all_planes, ax, alpha=0.5)
    # plot points
    ax.scatter(*scale_points(points), s=0.1, c='k')
    set_axes_equal(ax, ignore_z=True)
    ax.set_zlim3d([-4, 6])
    ax.view_init(elev=elev, azim=azim)
    print("Should see triangulation point cloud")
    fig.savefig("assets/scratch/Basic25DAlgorithm_mesh.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_mesh.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()

    print("")
    print("Should see two planes extracted, please rotate.")
    fig, ax = plt.subplots(figsize=(10, 10), nrows=1, ncols=1,
                           subplot_kw=dict(projection='3d'))
    # plot all triangles
    plot_polygons_3d(points, polygons, ax)
    plot_planes_3d(points, triangles, planes, ax)
    # plot points
    ax.scatter(*scale_points(points), c='k', s=0.1)
    set_axes_equal(ax)
    ax.view_init(elev=elev, azim=azim)
    fig.savefig("assets/scratch/Basic25DAlgorithm_polygons.pdf", bbox_inches='tight')
    fig.savefig("assets/scratch/Basic25DAlgorithm_polygons.png", bbox_inches='tight', pad_inches=-0.8)
    plt.show()
    print("")


if __name__ == "__main__":
    main()



""" Manage Versions in a python project
This will bump the file holding the current version (src/version.txt) as well as tag releases 
"""
import argparse
from pathlib import Path
from git import Repo
import sys

def parse_args():
    parser = argparse.ArgumentParser("Manage the project version including git tags.")
    parser.add_argument('--bump',
                        default='none',
                        const='none',
                        nargs='?',
                        choices=['major', 'minor', 'patch', 'none'],
                        help='Bump version of project')
    parser.add_argument('--version_file',
                        default='src/version.txt',
                        help='Path to version file')
    parser.add_argument("--tag",
                        dest="tag",
                        action="store_true",
                        default=False,
                        help="Create git tag")
    args = parser.parse_args()

    assert Path(args.version_file).exists()

    return args

def bump(bump_type:str='patch', version_file='src/version.txt'):
    with open('src/version.txt', 'r') as f:
        lines = f.readlines()
    version = ([int(line.split(' ')[1].strip()) for line in lines if line])

    # Increment the version, resetting minor/patches
    if bump_type == 'major':
        version[0] += 1
        version[1] = 0
        version[2] = 0
    if bump_type == 'minor':
        version[1] += 1
        version[2] = 0
    if bump_type == 'patch':
        version[2] += 1
    
    if bump_type != 'none':
        with open('src/version.txt', 'w') as f:
            f.write("MAJOR {}\nMINOR {}\nPATCH {}".format(*version))
    else:
        pass
        # print("Not incrementing version, bump was 'none'.\n")

    string_ints = [str(int_version) for int_version in version]
    version_str = ".".join(string_ints)

    return version_str

def create_tag(version_str, message="Tagged for release"):
    version_str = 'v{}'.format(version_str)
    repo = Repo('.')
    if repo.is_dirty():
        resp:str = input("Repo is dirty do you want to continue? [y/n]: ")
        if resp.lower() != 'y':
            print("Quiting...")
            sys.exit()
    hexsha = repo.head.commit.hexsha
    print("Creating new tag: {} ; commit: {}".format(version_str, hexsha))
    new_tag = repo.create_tag(version_str, message=message)
    return version_str
    
def main():
    args = parse_args()
    if args.tag:
        version_str = bump(bump_type='none', version_file=args.version_file)
        new_tag = create_tag(version_str)
        print("Don't forget to push the tags with:")
        print("\t git push origin {}  OR".format(new_tag))
        print("\t git push origin --tags")
    else:
        version_str = bump(bump_type=args.bump, version_file=args.version_file)
        print('Incremented version to {}'.format(version_str))
    



if __name__ == "__main__":
    main()


"""Paths and patches"""

from matplotlib.patches import PathPatch
from matplotlib.path import Path
from numpy import asarray, concatenate, ones


class Polygon(object):
    # Adapt Shapely or GeoJSON/geo_interface polygons to a common interface
    def __init__(self, context):
        if isinstance(context, dict):
            self.context = context['coordinates']
        else:
            self.context = context

    @property
    def exterior(self):
        return (getattr(self.context, 'exterior', None)
                or self.context[0])

    @property
    def interiors(self):
        value = getattr(self.context, 'interiors', None)
        if value is None:
            value = self.context[1:]
        return value


def PolygonPath(polygon):
    """Constructs a compound matplotlib path from a Shapely or GeoJSON-like
    geometric object"""

    def coding(ob):
        # The codes will be all "LINETO" commands, except for "MOVETO"s at the
        # beginning of each subpath
        n = len(getattr(ob, 'coords', None) or ob)
        vals = ones(n, dtype=Path.code_type) * Path.LINETO
        vals[0] = Path.MOVETO
        return vals

    if hasattr(polygon, 'geom_type'):  # Shapely
        ptype = polygon.geom_type
        if ptype == 'Polygon':
            polygon = [Polygon(polygon)]
        elif ptype == 'MultiPolygon':
            polygon = [Polygon(p) for p in polygon]
        else:
            raise ValueError(
                "A polygon or multi-polygon representation is required")

    else:  # GeoJSON
        polygon = getattr(polygon, '__geo_interface__', polygon)
        ptype = polygon["type"]
        if ptype == 'Polygon':
            polygon = [Polygon(polygon)]
        elif ptype == 'MultiPolygon':
            polygon = [Polygon(p) for p in polygon['coordinates']]
        else:
            raise ValueError(
                "A polygon or multi-polygon representation is required")

    vertices = concatenate([
        concatenate([asarray(t.exterior.coords)[:, :2]] +
                    [asarray(r)[:, :2] for r in t.interiors])
        for t in polygon])
    codes = concatenate([
        concatenate([coding(t.exterior.coords)] +
                    [coding(r) for r in t.interiors]) for t in polygon])

    return Path(vertices, codes)


def PolygonPatch(polygon, **kwargs):
    """Constructs a matplotlib patch from a geometric object

    The `polygon` may be a Shapely or GeoJSON-like object with or without holes.
    The `kwargs` are those supported by the matplotlib.patches.Polygon class
    constructor. Returns an instance of matplotlib.patches.PathPatch.

    Example (using Shapely Point and a matplotlib axes):

      >>> b = Point(0, 0).buffer(1.0)
      >>> patch = PolygonPatch(b, fc='blue', ec='blue', alpha=0.5)
      >>> axis.add_patch(patch)

    """
    return PathPatch(PolygonPath(polygon), **kwargs)


from os import path
import pytest
import numpy as np


from tests.python.helpers.utils import verify_points, basic_polylidar_verification, verify_all_polygons_are_valid, verify_all


np.random.seed(1)

def test_verify_building1(building1):
    verify_points(building1, 4)

def test_verify_building2(building2):
    verify_points(building2, 4)

def test_verify_hardcase1(hardcase1):
    verify_points(hardcase1, 2)

def test_building1(building1, basic_params):
    verify_all(building1, basic_params)


def test_building2(building2, basic_params):
    verify_all(building2, basic_params)

def test_hardcase1(hardcase1, hardcase1_params):
    verify_all(hardcase1, hardcase1_params)

# def test_bad_convex_hull(bad_convex_hull, bad_convex_hull_params):
#     verify_all(bad_convex_hull, bad_convex_hull_params)

def test_random_points(random_points, params_lmax):
    verify_all(random_points, params_lmax)

def test_clusters(cluster_groups):
    points = cluster_groups['points']
    polylidar_kwargs = cluster_groups['polylidar_kwargs']
    verify_all(points, polylidar_kwargs)





import time
import logging

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors

# from polylidar_plane_benchmark.utility.o3d_util import create_open_3d_mesh_from_tri_mesh

import organizedpointfilters as opf

from organizedpointfilters import Matrix3f, Matrix3fRef

from polylidar import extract_tri_mesh_from_organized_point_cloud, MatrixDouble

COLOR_PALETTE = list(map(colors.to_rgb, plt.rcParams['axes.prop_cycle'].by_key()['color']))


# def vector_magnitude(vec):
#     """
#     Calculates a vector's magnitude.
#     Args:
#         - vec ():
#     """
#     magnitude = np.sqrt(np.sum(vec**2))
#     return(magnitude)


# def align_vector_to_another(a=np.array([0, 0, 1]), b=np.array([1, 0, 0])):
#     """
#     Aligns vector a to vector b with axis angle rotation
#     """
#     if np.array_equal(a, b):
#         return None, None
#     axis_ = np.cross(a, b)
#     axis_ = axis_ / np.linalg.norm(axis_)
#     angle = np.arccos(np.dot(a, b))

#     return axis_, angle


# def create_arrow(scale=1, cylinder_radius=None, **kwargs):
#     """
#     Create an arrow in for Open3D
#     """
#     cone_height = scale * 0.2
#     cylinder_height = scale * 0.8
#     cone_radius = cylinder_radius if cylinder_radius else scale / 10
#     cylinder_radius = cylinder_radius if cylinder_radius else scale / 20
#     mesh_frame = o3d.geometry.TriangleMesh.create_arrow(cone_radius=cone_radius,
#                                                         cone_height=cone_height,
#                                                         cylinder_radius=cylinder_radius,
#                                                         cylinder_height=cylinder_height)
#     return(mesh_frame)


# def get_arrow(origin=[0, 0, 0], end=None, vec=None, **kwargs):
#     """
#     Creates an arrow from an origin point to an end point,
#     or create an arrow from a vector vec starting from origin.
#     Args:
#         - end (): End point. [x,y,z]
#         - vec (): Vector. [i,j,k]
#     """
#     # print(end)
#     scale = 10
#     beta = 0
#     gamma = 0
#     T = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
#     T[:3, -1] = origin
#     if end is not None:
#         vec = np.array(end) - np.array(origin)
#     elif vec is not None:
#         vec = np.array(vec)
#     if end is not None or vec is not None:
#         scale = vector_magnitude(vec)
#         mesh = create_arrow(scale, **kwargs)
#         axis, angle = align_vector_to_another(b=vec / scale)
#         if axis is None:
#             axis_a = axis
#         else:
#             axis_a = axis * angle
#             rotation_3x3 = mesh.get_rotation_matrix_from_axis_angle(axis_a)
#     # mesh.transform(T)
#     if axis is not None:
#         mesh = mesh.rotate(rotation_3x3, center=False)
#     mesh.translate(origin)
#     return(mesh)

def create_open_3d_mesh_from_tri_mesh(tri_mesh):
    """Create an Open3D Mesh given a Polylidar TriMesh"""
    triangles = np.asarray(tri_mesh.triangles)
    vertices = np.asarray(tri_mesh.vertices)
    triangle_normals = np.asarray(tri_mesh.triangle_normals)
    return create_open_3d_mesh(triangles, vertices, triangle_normals)


def create_open_3d_mesh(triangles, points, triangle_normals=None, color=COLOR_PALETTE[0], counter_clock_wise=True):
    """Create an Open3D Mesh given triangles vertices

    Arguments:
        triangles {ndarray} -- Triangles array
        points {ndarray} -- Points array

    Keyword Arguments:
        color {list} -- RGB COlor (default: {[1, 0, 0]})

    Returns:
        mesh -- Open3D Mesh
    """
    import open3d as o3d
    mesh_2d = o3d.geometry.TriangleMesh()
    if points.ndim == 1:
        points = points.reshape((int(points.shape[0] / 3), 3))
    if triangles.ndim == 1:
        triangles = triangles.reshape((int(triangles.shape[0] / 3), 3))
        # Open 3D expects triangles to be counter clockwise
    if not counter_clock_wise:
        triangles = np.ascontiguousarray(np.flip(triangles, 1))
    mesh_2d.triangles = o3d.utility.Vector3iVector(triangles)
    mesh_2d.vertices = o3d.utility.Vector3dVector(points)
    if triangle_normals is None:
        mesh_2d.compute_vertex_normals()
        mesh_2d.compute_triangle_normals()
    elif triangle_normals.ndim == 1:
        triangle_normals_ = triangle_normals.reshape((int(triangle_normals.shape[0] / 3), 3))
        mesh_2d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals_)
    else:
        mesh_2d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals)
    mesh_2d.paint_uniform_color(color)
    mesh_2d.compute_vertex_normals()
    return mesh_2d


def laplacian_opc(opc, loops=5, _lambda=0.5, kernel_size=3, **kwargs):
    """Performs Laplacian Smoothing on an organized point cloud

    Arguments:
        opc {ndarray} -- Organized Point Cloud MXNX3, Assumed F64

    Keyword Arguments:
        loops {int} -- How many iterations of smoothing (default: {5})
        _lambda {float} -- Weight factor for update (default: {0.5})
        kernel_size {int} -- Kernel Size (How many neighbors to intregrate) (default: {3})

    Returns:
        ndarray -- Smoothed Point Cloud, MXNX3, F64
    """
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    a_ref = Matrix3fRef(opc_float)

    t1 = time.perf_counter()
    if kernel_size == 3:
        b_cp = opf.filter.laplacian_K3(a_ref, _lambda=_lambda, iterations=loops, **kwargs)
    else:
        b_cp = opf.filter.laplacian_K5(a_ref, _lambda=_lambda, iterations=loops, **kwargs)
    t2 = time.perf_counter()
    logging.debug("OPC Mesh Smoothing Took (ms): %.2f", (t2 - t1) * 1000)

    opc_float_out = np.asarray(b_cp)
    opc_out = opc_float_out.astype(np.float64)

    time_elapsed = (t2 - t1) * 1000

    return opc_out, time_elapsed


def laplacian_opc_cuda(opc, loops=5, _lambda=0.5, kernel_size=3, **kwargs):
    """Performs Laplacian Smoothing on an organized point cloud

    Arguments:
        opc {ndarray} -- Organized Point Cloud MXNX3, Assumed F64

    Keyword Arguments:
        loops {int} -- How many iterations of smoothing (default: {5})
        _lambda {float} -- Weight factor for update (default: {0.5})
        kernel_size {int} -- Kernel Size (How many neighbors to intregrate) (default: {3})

    Returns:
        ndarray -- Smoothed Point Cloud, MXNX3
    """
    import organizedpointfilters.cuda as opf_cuda
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    t1 = time.perf_counter()
    if kernel_size == 3:
        opc_float_out = opf_cuda.kernel.laplacian_K3_cuda(opc_float, loops=loops, _lambda=_lambda, **kwargs)
    else:
        opc_float_out = opf_cuda.kernel.laplacian_K5_cuda(opc_float, loops=loops, _lambda=_lambda, **kwargs)
    t2 = time.perf_counter()

    logging.debug("OPC CUDA Laplacian Mesh Smoothing Took (ms): %.2f", (t2 - t1) * 1000)

    # only for visualization purposes here
    opc_out = opc_float_out.astype(np.float64)
    time_elapsed = (t2 - t1) * 1000

    return opc_out, time_elapsed

def compute_normals_and_centroids_opc(opc, convert_f64=True, **kwargs):
    """Computes the Normals and Centroid of Implicit Triangle Mesh in the Organized Point Cloud
    
    Arguments:
        opc {ndarray} -- MXNX3
    
    Keyword Arguments:
        convert_f64 {bool} -- Return F64? (default: {True})
    
    Returns:
        ndarray -- Numpy array
    """
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    a_ref = Matrix3fRef(opc_float)

    t1 = time.perf_counter()
    normals, centroids = opf.filter.compute_normals_and_centroids(a_ref)
    t2 = time.perf_counter()
    logging.debug("OPC Compute Normals and Centroids Took (ms): %.2f", (t2 - t1) * 1000)
    normals_float_out = np.asarray(normals)
    centroids_float_out = np.asarray(centroids)

    if not convert_f64:
        return (normals_float_out, centroids_float_out)

    return (normals_float_out.astype(np.float64), centroids_float_out.astype(np.float64))

def bilateral_opc(opc, loops=5, sigma_length=0.1, sigma_angle=0.261, **kwargs):
    """Performs bilateral normal smoothing on a mesh implicit from an organized point

    Arguments:
        opc {ndarray} -- Organized Point Cloud MXNX3, Assumed Float 64

    Keyword Arguments:
        loops {int} -- How many iterations of smoothing (default: {5})
        sigma_length {float} -- Saling factor for length (default: {0.1})
        sigma_angle {float} -- Scaling factor for angle (default: {0.261})

    Returns:
        ndarray -- MX3 Triangle Normal Array, Float 64
    """
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    a_ref = Matrix3fRef(opc_float)

    t1 = time.perf_counter()
    normals = opf.filter.bilateral_K3(a_ref, iterations=loops, sigma_length=sigma_length, sigma_angle=sigma_angle)
    t2 = time.perf_counter()
    logging.debug("OPC Bilateral Filter Took (ms): %.2f", (t2 - t1) * 1000)
    normals_float_out = np.asarray(normals)
    normals_out = normals_float_out.astype(np.float64)

    time_elapsed = (t2-t1) * 1000

    return normals_out, time_elapsed

def bilateral_opc_cuda(opc, loops=5, sigma_length=0.1, sigma_angle=0.261, **kwargs):
    """Performs bilateral normal smoothing on a mesh implicit from an organized point

    Arguments:
        opc {ndarray} -- Organized Point Cloud MXNX3, Assumed Float 64

    Keyword Arguments:
        loops {int} -- How many iterations of smoothing (default: {5})
        sigma_length {float} -- Saling factor for length (default: {0.1})
        sigma_angle {float} -- Scaling factor for angle (default: {0.261})

    Returns:
        ndarray -- MX3 Triangle Normal Array, Float 64
    """
    import organizedpointfilters.cuda as opf_cuda
    normals_opc, centroids_opc = compute_normals_and_centroids_opc(opc, convert_f64=False)
    assert normals_opc.dtype == np.float32
    assert centroids_opc.dtype == np.float32

    t1 = time.perf_counter()
    normals_float_out = opf_cuda.kernel.bilateral_K3_cuda(
        normals_opc, centroids_opc, loops=loops, sigma_length=sigma_length, sigma_angle=sigma_angle)
    t2 = time.perf_counter()

    normals_out = normals_float_out.astype(np.float64)

    time_elapsed = (t2-t1) * 1000

    logging.debug("OPC CUDA Bilateral Filter Took (ms): %.2f", (t2 - t1) * 1000)

    return normals_out, time_elapsed


def laplacian_then_bilateral_opc(opc, loops_laplacian=5, _lambda=1.0, kernel_size=3, loops_bilateral=0, sigma_length=0.1, sigma_angle=0.261, **kwargs):
    """Performs Laplacian Smoothing on Point Cloud and then performs Bilateral normal smoothing

    Arguments:
        opc {ndarray} -- Organized Point Cloud (MXNX3)

    Keyword Arguments:
        loops_laplacian {int} -- How many iterations of laplacian smoothing (default: {5})
        _lambda {float} -- Weigh factor for laplacian  (default: {0.5})
        kernel_size {int} -- Kernel Size for Laplacian (default: {3})
        loops_bilateral {int} -- How many iterations of bilateral smoothing (default: {0})
        sigma_length {float} -- Scaling factor for length bilateral (default: {0.1})
        sigma_angle {float} -- Scaling factor for angle bilateral (default: {0.261})

    Returns:
        tuple(ndarray, ndarray) -- Smoothed OPC MXNX3, Smoothed Normals M*NX3, normals arrays are flattened
    """
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    a_ref = Matrix3fRef(opc_float)

    t1 = time.perf_counter()
    if kernel_size == 3:
        b_cp = opf.filter.laplacian_K3(a_ref, _lambda=_lambda, iterations=loops_laplacian, **kwargs)
    else:
        b_cp = opf.filter.laplacian_K5(a_ref, _lambda=_lambda, iterations=loops_laplacian, **kwargs)
    t2 = time.perf_counter()
    logging.debug("OPC Mesh Smoothing Took (ms): %.2f", (t2 - t1) * 1000)

    opc_float_out = np.asarray(b_cp)
    b_ref = Matrix3fRef(opc_float_out)

    opc_normals_float = opf.filter.bilateral_K3(
        b_ref, iterations=loops_bilateral, sigma_length=sigma_length, sigma_angle=sigma_angle)
    t3 = time.perf_counter()
    logging.debug("OPC Bilateral Normal Filter Took (ms): %.2f", (t3 - t2) * 1000)

    opc_normals_float_out = np.asarray(opc_normals_float)
    total_triangles = int(opc_normals_float_out.size / 3)
    opc_normals_out = opc_normals_float_out.astype(np.float64)
    opc_normals_out = opc_normals_float_out.reshape((total_triangles, 3))

    opc_out = opc_float_out.astype(np.float64)
    # total_points = int(opc_out.size / 3)
    # opc_out = opc_out.reshape((total_points, 3))

    timings = dict(t_laplacian=(t2-t1)*1000, t_bilateral=(t3-t2)*1000)

    return opc_out, opc_normals_out, timings

def laplacian_then_bilateral_opc_cuda(opc, loops_laplacian=5, _lambda=1.0, kernel_size=3, loops_bilateral=0, sigma_length=0.1, sigma_angle=0.261, **kwargs):
    """Performs Laplacian Smoothing on Point Cloud and then performs Bilateral normal smoothing

    Arguments:
        opc {ndarray} -- Organized Point Cloud (MXNX3)

    Keyword Arguments:
        loops_laplacian {int} -- How many iterations of laplacian smoothing (default: {5})
        _lambda {float} -- Weigh factor for laplacian  (default: {0.5})
        kernel_size {int} -- Kernel Size for Laplacian (default: {3})
        loops_bilateral {int} -- How many iterations of bilateral smoothing (default: {0})
        sigma_length {float} -- Scaling factor for length bilateral (default: {0.1})
        sigma_angle {float} -- Scaling factor for angle bilateral (default: {0.261})

    Returns:
        tuple(ndarray, ndarray) -- Smoothed OPC MXNX3, Smoothed Normals M*NX3, the arrays are flattened
    """
    import organizedpointfilters.cuda as opf_cuda
    opc_float = (np.ascontiguousarray(opc[:, :, :3])).astype(np.float32)

    t1 = time.perf_counter()
    if kernel_size == 3:
        opc_float_out = opf_cuda.kernel.laplacian_K3_cuda(opc_float, loops=loops_laplacian, _lambda=_lambda)
    else:
        opc_float_out = opf_cuda.kernel.laplacian_K5_cuda(opc_float, loops=loops_laplacian, _lambda=_lambda)
    t2 = time.perf_counter()

    opc_normals, time_bilateral = bilateral_opc_cuda(opc_float_out, loops=loops_bilateral,
                                     sigma_length=sigma_length, sigma_angle=sigma_angle)

    opc_out = opc_float_out.astype(np.float64)
    # total_points = int(opc_out.size / 3)
    # opc_out = opc_out.reshape((total_points, 3))

    total_triangles = int(opc_normals.size / 3)
    opc_normals = opc_normals.reshape((total_triangles, 3))

    timings = dict(t_laplacian=(t2-t1)*1000, t_bilateral=time_bilateral)

    return opc_out, opc_normals, timings

def create_mesh_from_organized_point_cloud(pcd, rows=500, cols=500, stride=2, calc_normals=True):
    """Create Mesh from organized point cloud
    If an MXNX3 Point Cloud is passed, rows and cols is ignored (we know the row/col from shape)
    If an KX3 Point Cloud is passed, you must pass the row, cols, and stride that correspond to the point cloud

    Arguments:
        pcd {ndarray} -- Numpy array. Either a K X 3 (flattened) or MXNX3

    Keyword Arguments:
        rows {int} -- Number of rows (default: {500})
        cols {int} -- Number of columns (default: {500})
        stride {int} -- Stride used in creating point cloud (default: {2})

    Returns:
        tuple -- Polylidar Tri Mesh and O3D mesh
    """
    pcd_ = pcd
    if pcd.ndim == 3:
        rows = pcd.shape[0]
        cols = pcd.shape[1]
        stride = 1
        pcd_ = pcd.reshape((rows * cols, 3))

    pcd_mat = MatrixDouble(pcd_, copy=True)
    t1 = time.perf_counter()
    tri_mesh, tri_map = extract_tri_mesh_from_organized_point_cloud(pcd_mat, rows, cols, stride, calc_normals=calc_normals)
    t2 = time.perf_counter()
    time_elapsed = (t2 - t1) * 1000
    return tri_mesh, tri_map, time_elapsed

def plot_triangle_normals(normals:np.ndarray, normals2:np.ndarray):
    f, (ax1, ax2) = plt.subplots(1,2) 
    colors = ((normals * 0.5 + 0.5) * 255).astype(np.uint8)
    im = colors.reshape((249, 249, 2, 3))
    im = im[:, :, 1, :]

    colors2 = ((normals2 * 0.5 + 0.5) * 255).astype(np.uint8)
    im2= colors2.reshape((249, 249, 2, 3))
    im2 = im2[:, :, 1, :]

    ax1.imshow(im, origin='upper')
    ax2.imshow(im2, origin='upper')
    plt.show()


def pick_valid_normals(tri_map, opc_normals):
    tri_map_np = np.asarray(tri_map)
    mask = tri_map_np != np.iinfo(tri_map_np.dtype).max
    tri_norms = np.ascontiguousarray(opc_normals[mask,:])
    return tri_norms


def create_meshes(opc, **kwargs):
    """Creates a mesh from a noisy organized point cloud

    Arguments:
        opc {ndarray} -- Must be MXNX3

    Keyword Arguments:
        loops {int} -- How many loop iterations (default: {5})
        _lambda {float} -- weighted iteration movement (default: {0.5})

    Returns:
        [tuple(mesh, o3d_mesh)] -- polylidar mesh and o3d mesh reperesentation
    """
    smooth_opc, opc_normals, timings = laplacian_then_bilateral_opc(opc, **kwargs)
    tri_mesh, tri_map, time_elapsed_mesh = create_mesh_from_organized_point_cloud(smooth_opc, calc_normals=False)
    tri_norms = pick_valid_normals(tri_map, opc_normals)
    opc_normals_cp = MatrixDouble(tri_norms, copy=True) # copy here!!!!!
    # plot_triangle_normals(np.asarray(tri_mesh.triangle_normals), opc_normals)
    tri_mesh.set_triangle_normals(opc_normals_cp) # copy again here....sad
    timings = dict(**timings, t_mesh=time_elapsed_mesh)
    return tri_mesh, timings

def create_meshes_cuda(opc, **kwargs):
    """Creates a mesh from a noisy organized point cloud

    Arguments:
        opc {ndarray} -- Must be MXNX3

    Keyword Arguments:
        loops {int} -- How many loop iterations (default: {5})
        _lambda {float} -- weighted iteration movement (default: {0.5})

    Returns:
        [tuple(mesh, o3d_mesh)] -- polylidar mesh and o3d mesh reperesentation
    """
    smooth_opc, opc_normals, timings = laplacian_then_bilateral_opc_cuda(opc, **kwargs)
    tri_mesh, tri_map, time_elapsed_mesh = create_mesh_from_organized_point_cloud(smooth_opc, calc_normals=False)
    tri_norms = pick_valid_normals(tri_map, opc_normals)
    opc_normals_cp = MatrixDouble(tri_norms, copy=True) # copy here!!!!!
    # plot_triangle_normals(np.asarray(tri_mesh.triangle_normals), opc_normals)
    tri_mesh.set_triangle_normals(opc_normals_cp) # copy again here....sad
    timings = dict(**timings, t_mesh=time_elapsed_mesh)
    return tri_mesh, timings

def create_meshes_cuda_with_o3d(opc, **kwargs):
    """Creates a mesh from a noisy organized point cloud

    Arguments:
        opc {ndarray} -- Must be MXNX3

    Keyword Arguments:
        loops {int} -- How many loop iterations (default: {5})
        _lambda {float} -- weighted iteration movement (default: {0.5})

    Returns:
        [tuple(mesh, o3d_mesh)] -- polylidar mesh and o3d mesh reperesentation
    """
    tri_mesh, timings = create_meshes_cuda(opc, **kwargs)
    tri_mesh_o3d = create_open_3d_mesh_from_tri_mesh(tri_mesh)

    return tri_mesh, tri_mesh_o3d, timings
    

